# Copyright (C) 2016 Red Hat, Inc.
# This file is part of libsan.
#
# libsan is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# libsan is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with libsan.  If not, see <http://www.gnu.org/licenses/>.

"""net.py: Module to manipulate network devices."""

import ipaddress
import os.path
import re  # regex
from os import listdir, readlink
from os.path import join, lexists

import netifaces

import libsan.host.linux
from libsan import _print
from libsan.host.cmdline import exists, run

__author__ = "Bruno Goncalves"
__copyright__ = "Copyright (c) 2016 Red Hat, Inc. All rights reserved."

sysfs_class_net_path = "/sys/class/net"


def is_mac(mac):
    """Check if given MAC is  on valid format"""
    if standardize_mac(mac):
        return True
    return False


def get_nics():
    """Return list of all NICs on the server"""
    ifaces = netifaces.interfaces()
    if ifaces is None:  # No NIC on this server
        return None

    return ifaces


def get_mac_of_nic(nic):
    """Given a NIC name return its MAC address"""

    try:
        mac = netifaces.ifaddresses(nic)[netifaces.AF_PACKET][0]["addr"]
    except Exception as e:
        print(repr(e))
        return None
    else:
        return mac


def get_nic_of_mac(mac_address):
    """Give an MAC address return the server interface name"""
    if not mac_address:
        _print("FAIL: get_nic_of_mac() - requires mac as argument")
        return None

    mac = mac_address.lower()

    if not is_mac(mac):
        _print("FAIL: get_nic_of_mac() - %s does not seem to be a valid MAC" % mac)
        return None

    nics = get_nics()
    if not nics:
        return None

    for nic in nics:
        if mac == get_mac_of_nic(nic):
            return nic

    return None


def get_ip_address_of_nic(nic):
    """Get IPv4 of specific network interface"""
    try:
        ip = netifaces.ifaddresses(nic)[netifaces.AF_INET][0]["addr"]
    except KeyError as e:
        print("KeyError - Iface probably does not have a IP address - %s" % str(e))
    except Exception as e:
        print(repr(e))
    else:
        return ip
    return None


def get_ipv6_address_of_nic(nic):
    """Get IPv6 of specific network interface"""
    try:
        ip = netifaces.ifaddresses(nic)[netifaces.AF_INET6][0]["addr"]
    except KeyError as e:
        print("KeyError - Iface probably does not have a IPv6 address - %s" % str(e))
    except Exception as e:
        print(repr(e))
    else:
        return ip
    return None


def get_nic_of_ip(ip):
    """Given an IP address return the NIC name using it"""
    if not ip:
        return None

    nics = get_nics()
    if not nics:
        return None

    for nic in nics:
        if ip == get_ip_address_of_nic(nic):
            return nic
    return None


def nic_2_driver():
    """Return a dictionary where nic name is the key and driver name is the value.
    Will skip sub interfaces, loop device, tun, vboxnet0.
    The arguments are:
    \tNone
    \treturn_output (Dict): Return a dictionary
    Returns:
    \tdict: Return dict containing all NICs"""
    nic_dict = {}
    for nic in listdir(sysfs_class_net_path):
        if (
            nic == "."
            or nic == ".."
            or nic == "lo"
            or re.match("^tun[0-9]+", nic)  # loop
            or re.match("^vboxnet[0-9]+", nic)  # TUN NIC
            or re.search(r"\.", nic)  # virtualbox NIC.
        ):  # sub-interface
            continue
        nic_dict[nic] = driver_of_nic(nic)
    # print nic_dict
    return nic_dict


# End of nic_2_driver()


def driver_of_nic(nic):
    """Given a specific NIC name it returns its driver name
    Find out the driver of certain NIC via sysfs file:
        /sys/class/net/eth0/device/driver   # it's a link.
    The arguments are:
    \tnic: NIC name, e.g. eth0
    Returns:
    \tstr: Driver name"""
    nic = phy_nic_of(nic)
    nic_file = join(sysfs_class_net_path, nic)
    if not lexists(nic_file):
        print("FAIL: No such NIC exists: %s" % nic)
        print(nic_file)
        return None
    driver_file = "/sys/class/net/%s/device/driver" % nic
    if not lexists(driver_file):
        print("FAIL: path %s does not exist" % driver_file)
        return None
    # from a symlink get real path
    real_path = os.readlink(driver_file)
    m = re.match(".*drivers/(.*)$", real_path)
    if not m:
        print("FAIL: Could not find driver name for %s" % nic)
        return None
    return m.group(1)


# End of driver_of_nic()


def phy_nic_of(nic):
    """Translate sub-interface of NIC 'eth0.802-fcoe' to physical NIC 'eth0'.
    The arguments are:
    \tnic: NIC name, e.g. eth0.802-fcoe
    Returns:
    \tstr: phy NIC, e.g. eth0
    """
    if not nic:
        return None
    phy_nic = re.sub(r"\..+$", "", nic)
    return phy_nic


def get_pci_id_of_nic(nic):
    """From a specific network interface return its PCI ID"""
    regex_pci_id = libsan.host.linux.get_regex_pci_id()
    sys_path = f"{sysfs_class_net_path}/{nic}"
    link_path = readlink(sys_path)
    # _print("DEBUG: get_pci_id_of_nic - %s" % link_path)
    m = re.search(f"({regex_pci_id})/net/{nic}", link_path)
    if m:
        return m.group(1)
    return None


def get_ip_version(addr):
    """Given an address, tries to check if it is IPv6 or not
    The arguments are:
    \taddr:     Network address
    Returns:
    \t4:        If it is a valid IPv4 address
    \t6:        If it is a valid IPv6 address
    \tNone:     addr is not an IPv4 or IPv6 address
    """
    try:
        ipver = ipaddress.ip_address(addr).version
    except Exception as e:
        print(repr(e))
        return None
    else:
        return ipver


def standardize_mac(mac):
    """Usage
        standardize_mac(mac)
    Purpose
        Convert all possiable format mac into stand type:
            (?:[0-9A-F]{2}:){5}[0-9A-F]{2} #like: F0:DE:F1:0D:D3:C9
        Return STRING or ARRAY base on context.
    Parameter
        mac           # any format mac, like "0005.73dd.9a19"
    Returns
        mac
            or
        None
    """
    if not mac:
        return None
    regex_standard_mac = "^(?:[0-9A-F]{2}:){5}[0-9A-F]{2}$"

    mac = mac.lower()
    mac = re.sub("^0x", "", mac)
    mac = re.sub("[^0-9A-Fa-f]", "", mac)
    # If mac given has no ':' we will add it
    if re.match("[0-9a-f]{12}", mac):
        mac_regex = re.compile("(.{2})")
        mac = mac_regex.sub(r"\g<1>:", mac)
        mac = re.sub(":$", "", mac)

    if re.match(regex_standard_mac, mac, re.IGNORECASE):
        return mac

    return None


def convert_netmask(netmask="255.255.255.0"):
    """Converts subnet mask to CIDR prefix.
    \t netmask: common subnet mask
    """
    try:
        cidr = ipaddress.IPv4Network((0, netmask)).prefixlen
    except ValueError as e:
        print(e)
        return None
    else:
        return cidr


def if_down(nic_or_mac):
    """Bring the interface down using ifdown tool
    Parameters:
    \t Interface name or it's MAC address
    """
    if not nic_or_mac:
        _print("FAIL: if_down() - requires nic or mac as argument")
        return None

    # ifup/ifdown scripts are not shipped by default on RHEL-8+
    if not exists("ifdown") and libsan.host.linux.is_installed("NetworkManager"):  # noqaSIM102
        if not libsan.host.linux.install_package("NetworkManager-initscripts-updown"):
            print('FAIL: unable to run "ifdown" command')
            return False

    nic = get_nic_of_mac(nic_or_mac) if is_mac(nic_or_mac) else nic_or_mac

    if run("ifdown %s" % nic):
        return True
    return False


def if_up(nic_or_mac):
    """Bring the interface up using ifup tool
    Parameters:
    \t Interface name or it's MAC address
    """
    if not nic_or_mac:
        _print("FAIL: if_up() - requires nic or mac as argument")
        return None

    # ifup/ifdown scripts are not shipped by default on RHEL-8+
    if not exists("ifup") and libsan.host.linux.is_installed("NetworkManager"):  # noqa: SIM102
        if not libsan.host.linux.install_package("NetworkManager-initscripts-updown"):
            print('FAIL: unable to run "ifup" command')
            return False

    nic = get_nic_of_mac(nic_or_mac) if is_mac(nic_or_mac) else nic_or_mac

    if run("ifup %s" % nic):
        return True
    return False


def iface_up(nic_or_mac):
    """Bring the interface up
    Parameters:
    \t Interface name or it's MAC address
    """
    if not nic_or_mac:
        _print("FAIL: iface_up() - requires nic or mac as argument")
        return False

    nic = get_nic_of_mac(nic_or_mac) if is_mac(nic_or_mac) else nic_or_mac

    retcode, output = run("ip link set %s up" % nic, return_output=True)
    if retcode != 0:
        print(output)
        return False
    return True


def iface_down(nic_or_mac):
    """Bring the interface down
    Parameters:
    \t Interface name or it's MAC address
    """
    if not nic_or_mac:
        _print("FAIL: iface_down() - requires nic or mac as argument")
        return False

    nic = get_nic_of_mac(nic_or_mac) if is_mac(nic_or_mac) else nic_or_mac

    retcode, output = run("ip link set %s down" % nic, return_output=True)
    if retcode != 0:
        print(output)
        return False
    return True


def set_ifcfg(nic_or_mac, parameters):
    """Edit or create ifcfg files: IP, prefix, gateway, defroute...
    Parameters:
    \t nic_or_mac: interface name or mac address
    \t dict of params: e.g. {'IPADDR': '10.37.151.7'}
    """

    if is_mac(nic_or_mac):
        nic = get_nic_of_mac(nic_or_mac)
        if not nic:
            print("FAIL: Couldn't find NIC from MAC.")
            return False
    else:
        nic = nic_or_mac

    file_to_edit = f"/etc/sysconfig/network-scripts/ifcfg-{nic}"
    if not libsan.host.linux.edit_config(file_to_edit, parameters):
        return False

    if_down(nic)
    if_up(nic)
    return True


def get_default_iface():
    """Returns tuple with ip and interface.
    example: ('10.1.1.1', 'eno1')
    """
    try:
        return netifaces.default_gateway()[netifaces.AF_INET]
    except Exception as e:
        print(repr(e))
        return None


def nm_get_conn(nic_or_mac):
    """Returns NetworkManager connection UUID
    \t nic_or_mac: interface name or mac address
    """
    if is_mac(nic_or_mac):
        nic = get_nic_of_mac(nic_or_mac)
        if not nic:
            print("FAIL: Couldn't find NIC from MAC.")
            return None
    else:
        nic = nic_or_mac

    conn = nm_get_conn_from_dev(nic)
    if not conn:
        print("WARN: %s not used in any active connection. Trying connections with same name" % nic)
        # Check connection with same name as device
        if nm_get_conn_iface(nic) == nic:
            return nm_get_conn_uuid(nic)
        return None
    conn = str(conn)
    return nm_get_conn_uuid(conn)


def nm_get_conn_iface(conn):
    """Returns interface name used by NM connection
    \t conn: connection id or uuid
    """
    cmd = "nmcli -g connection.interface-name con show '%s'" % conn
    retcode, output = run(cmd, return_output=True, verbose=False)
    if retcode != 0:
        print("FAIL: Couldn't get connection.interface-name for connection " + conn)
        print(output)
        return None
    return output


def nm_get_conn_uuid(conn):
    """Returns NetworkManager connection UUID
    \t conn: connection id
    """
    cmd = "nmcli -g connection.uuid conn show '%s'" % conn
    print(cmd)
    retcode, output = run(cmd, return_output=True, verbose=False)
    if retcode != 0:
        print("FAIL: Couldn't get NM connection uuid using " + conn)
        print(output)
        return None
    return output


def nm_get_conn_from_dev(nic):
    """Returns connection id(name) using specified device.
    \t nic: network interface - device
    """
    nic = str(nic)
    cmd = "nmcli -g GENERAL.CONNECTION device show " + nic

    retcode, output = run(cmd, return_output=True, verbose=False)
    if retcode != 0:
        print("FAIL: Couldn't get NM connection using device " + nic)
        print(output)
        return None
    return output


def nm_get_dev_from_conn(conn):
    """Returns a device used by specified connection
    \t conn: networkmanager connection id(name) or uuid
    """
    conn = str(conn)
    cmd = "nmcli -g connection.interface-name con show '%s'" % conn

    retcode, output = run(cmd, return_output=True, verbose=False)
    if retcode != 0:
        print("FAIL: Couldn't get device used by NM connection " + conn)
        print(output)
        return None
    return output


def nm_conn_up(conn):
    """Uses nmcli to activate connection
    \t conn: connection id(name) or uuid
    """
    cmd = 'nmcli conn up "%s"' % conn
    retcode, output = run(cmd, return_output=True, verbose=False)
    if retcode != 0:
        print("FAIL: Unable to activate the connection: %s" % conn)
        return False
    print(output)
    return True


def nm_conn_down(conn):
    """Uses nmcli to deactivate connection
    \t conn: connection id(name) or uuid
    """
    cmd = 'nmcli conn down "%s"' % conn
    retcode, output = run(cmd, return_output=True, verbose=False)
    if retcode != 0:
        print("FAIL: Unable to deactivate the connection: %s" % conn)
        return False
    print(output)
    return True


def nm_conn_reload():
    """Runs `nmcli conn reload`. Does not support RHEL6"""
    cmd = "nmcli conn reload"
    retcode = run(cmd, verbose=False)
    if retcode != 0:
        print("FAIL: Unable to reload NetworkManager")
        return False
    return True


def nm_conn_show(conn):
    """Use nmcli conn to show all connection parameters. Does not support RHEL6
    \t conn: networkmanager connection id(name) or uuid
    """
    if not conn:
        print("FAIL: No conn specified")
        return False

    cmd = 'nmcli conn show "%s"' % conn
    retcode, output = run(cmd, return_output=True, verbose=False)
    if retcode != 0:
        print("FAIL: Unable to show conn")
        return False
    print(output)
    return True


def nm_conn_del(conn):
    """Deletes NetworkManager connection using nmcli
    \t conn: connection id(name) or uuid
    """
    cmd = "nmcli conn delete %s" % conn
    retcode, output = run(cmd, return_output=True, verbose=False)
    if retcode != 0:
        print("FAIL: Unable to delete the conn %s" % conn)
        return False
    print(output)
    return True


def nm_add_conn(name=None, nic_or_mac=None, nic_type="ethernet"):
    """Use it to add new connection for devices without one
    \tcmd="nmcli con add type ethernet ifname enp5s0f1 con-name enp5s0f1 "
    """

    if is_mac(nic_or_mac):
        nic = get_nic_of_mac(nic_or_mac)
        if not nic:
            print("FAIL: Couldn't find NIC from MAC.")
            return False
    else:
        nic = nic_or_mac

    if not name:
        name = nic

    conn = None
    if nic:
        conn = nm_get_conn_from_dev(nic)
    # Exit if there is a connection already
    if conn:
        print("Connection already exists. Linked to: %s " % conn)
        return True

    cmd = f"nmcli conn add type {nic_type} con-name {name}"
    if nic:
        cmd += " ifname %s" % nic
    if "_" in cmd:
        cmd = cmd.replace("_", "-")
    retcode, output = run(cmd, return_output=True, verbose=False)
    if retcode != 0:
        print("FAIL: Unable to add con with cmd %s " % cmd)
        print(output)
        return False
    nm_conn_reload()
    return True


def nm_conn_mod(conn, key, value):
    """Modify one or more properties of the NetworkManager connection profile.
    \tnm_con_mod("enp17s0f1","connection.autoconnect", "yes")
    \tnmcli c modify ens2f1 connection.autoconnect yes
    \tExamples compared to network-scripts:
    \tipv4.method manual     >> BOOTPROTO=none
    \tipv4.method auto       >> BOOTPROTO=dhcp
    \tipv4.address "192.168.0.10/24"   >> IPADDR=192.168.0.10
    \tipv4.gateway 192.168.0.1  >> GATEWAY=192.168.0.1
    \tipv4.dns 8.8.8.8   >> DNS1=8.8.8.8
    \tconnection.autoconnect yes   >> ONBOOT=yes
    \tconnection.id eth0   >> NAME=eth0
    \tconnection.interface-name eth0   >> DEVICE=eth0
    \t802-3-ethernet.mac-address 08:00:27:4b:7a:80 >> HWADDR=08:00:27:4b:7a:80
    \tipv4.never-default no   >> DEFROUTE=yes
    """

    if not conn:
        print("FAIL: No connection specified")
        return False
    conn_uuid = nm_get_conn_uuid(conn)
    if not conn_uuid:
        return False
    cmd = "nmcli conn modify %s" % conn_uuid

    cmd += f" {key} {value}"

    retcode, output = run(cmd, return_output=True, verbose=False)
    if retcode != 0:
        print(f"FAIL: Unable to modify conn {conn} with cmd {cmd} ")
        print(output)
        return False
    return True


def nm_dev_mod(dev, key, value):
    """Modify one or more properties that are currently active on the device without modifying
    the connection profile. The changes have immediate effect.
    nmcli dev modify em1 ipv4.method shared
    nmcli dev modify em1 ipv4.address xx
    """

    cmd = "nmcli device modify %s" % dev
    cmd += f" {key} {value}"
    retcode, output = run(cmd, return_output=True, verbose=False)
    if retcode != 0:
        print(f"FAIL: Unable to modify dev {dev} with cmd {cmd} ")
        print(output)
        return False
    return True


def nm_set_ip(conn, ip, netmask="24", activate=True):
    """Uses nmcli to set static IP, netmask and activates connection.
    \t conn: networkmanager connection id(name) or uuid
    \t ip: IPv4 or IPv6
    """
    ip = str(ip)
    ipver = get_ip_version(ip)

    if ipver == 4:
        cmd_ip = "ipv4"
    elif ipver == 6:
        cmd_ip = "ipv6"
    else:
        print("FAIL: Invalid IP format")
        return False

    if "." in netmask:
        netmask = convert_netmask(netmask)

    cmd = f"nmcli conn modify {conn} {cmd_ip}.method manual {cmd_ip}.addr {ip}/{netmask}"
    retcode, output = run(cmd, return_output=True, verbose=False)
    if retcode != 0:
        print(f"FAIL: Unable to set IP using nmcli: {conn} {ip}")
        print(output)
        return False

    if activate and not nm_conn_up(conn):
        return False
    return True



#!/usr/bin/python
import pytest

# Copyright (C) 2016 Red Hat, Inc.
# This file is part of libsan.
#
# libsan is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# libsan is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with libsan.  If not, see <http://www.gnu.org/licenses/>.
from libsan.host import loopdev


def test_loopdev():
    dev = loopdev.create_loopdev()
    if not dev:
        print("SKIP: Could not create loop device")
        return

    if not loopdev.delete_loopdev(dev):
        pytest.fail("FAIL: Could not delete loop device")

    assert 1



#!/usr/bin/python

# Copyright (C) 2016 Red Hat, Inc.
# This file is part of libsan.
#
# libsan is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# libsan is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with libsan.  If not, see <http://www.gnu.org/licenses/>.

import unittest

import pytest
from libsan.host import ssh
from libsan.host.cmdline import run
from libsan.host.linux import is_docker

host = "localhost"
user = "root"
passwd = "redhat"
command = "uname -r"


def connect():
    return ssh.connect(host=host, user=user, passwd=passwd, port=22)


class TestSSH(unittest.TestCase):
    def test_connect(self):
        if not is_docker():
            assert connect() is not None

    def test_connect_no_passwd(self):
        assert ssh.connect(host=host, user=user, passwd="") is None

    def test_connect_passwd_is_none(self):
        assert ssh.connect(host=host, user=user, passwd=None) is None

    def test_connect_no_user(self):
        assert ssh.connect(host=host, user="", passwd=passwd) is None

    def test_connect_user_is_none(self):
        assert ssh.connect(host=host, user=None, passwd=passwd) is None

    def test_connect_no_host(self):
        assert ssh.connect(host="", user=user, passwd=passwd, max_attempt=1) is None

    def test_disconnect(self):
        if not is_docker():
            assert ssh.disconnect(connect()) is True

    def test_disconnect_none(self):
        with pytest.raises(Exception) as context:  # noqaPT011
            ssh.disconnect(None)
        assert "'NoneType' object has no attribute" in str(context)

    def test_command(self):
        if not is_docker():
            assert ssh.run_cmd(connect(), command) == 0

    def test_command_output(self):
        if not is_docker():
            command_run = run(command, return_output=True)
            command_ssh = ssh.run_cmd(connect(), command, return_output=True)
            assert command_run[1] + "\n" == command_ssh[1]

    def test_command_invoke_shell(self):
        if not is_docker():
            assert ssh.run_cmd(connect(), command, invoke_shell=True, expect="# ") == 0



# Copyright (C) 2016 Red Hat, Inc.
# This file is part of libsan.
#
# libsan is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# libsan is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with libsan.  If not, see <http://www.gnu.org/licenses/>.

"""mp.py: Module to manage multipath devices."""

__author__ = "Bruno Goncalves"
__copyright__ = "Copyright (c) 2016 Red Hat, Inc. All rights reserved."

import os.path
import re

import libsan.host.fc
import libsan.host.iscsi
import libsan.host.linux
import libsan.host.lvm
import libsan.host.net
import libsan.host.scsi
import libsan.misc.size
from libsan import _print
from libsan.host.cmdline import run

multipath_conf_path = "/etc/multipath.conf"
aug_conf_path = "/files/etc/multipath.conf"
package_name = "device-mapper-multipath"


def mp_service_name():
    """Return the name of mutlipath service."""
    return "multipathd"


def mp_start_service():
    """Start multipath service."""
    if libsan.host.linux.is_installed(package_name) and not os.path.isfile(multipath_conf_path):
        mpathconf_enable()
    return libsan.host.linux.service_start(mp_service_name())


def mp_stop_service():
    """Stop multipath service."""
    return libsan.host.linux.service_stop(mp_service_name())


def is_multipathd_running():
    """Check if multipathd is running."""
    return libsan.host.linux.is_service_running(mp_service_name())


def mpathconf_enable(find_mpaths=False) -> bool:
    """Runs 'mpathconf --enable' command"""
    cmd = "mpathconf --enable"
    if find_mpaths:
        cmd = cmd + " --find_multipaths n"
    retcode = run(cmd)
    if retcode != 0:
        return False

    # bugzilla.redhat.com/show_bug.cgi?id=2179859
    bad_line = 'enable_foreign "^$"'
    with open(multipath_conf_path, "r+") as f:
        lines = f.readlines()
        f.seek(0)
        for line in lines:
            if bad_line not in line:
                f.write(line)
            else:
                _print(f"INFO: Removing '{bad_line}' line from {multipath_conf_path}")
        f.truncate()

    return True


def mp_enable() -> bool:
    """Installs and enable multipath."""
    if not is_multipathd_running():
        libsan.host.linux.install_package("device-mapper-multipath")
        mpathconf_enable()
        if not libsan.host.linux.service_start(mp_service_name()):
            return False

    return True


def is_mpath_device(mpath_name, print_fail=True):
    """Checks if given device is multipath.
    :param mpath_name: name of device to check
    :param print_fail: Should we print "FAIL: ..." in case of fail?
    :return: True / False.
    """
    cmd = "multipath -l %s" % mpath_name
    retcode, output = run(cmd, return_output=True, verbose=False)
    if retcode != 0:
        if print_fail:
            _print('FAIL: Could not execute "%s"' % cmd)
        return None

    # There could be error in the output, so we need to check if we really found the mpath device.
    # 'multipath -l device_name' always should return device_name at the beginning if found
    if output and len(output) > 0 and output.startswith("%s " % mpath_name):
        return True
    return False


def get_mpath_of_disk(disk):
    """Get the multipath device of a disk
    The arguments are:
    \tDisk:   Disk
    Returns:
    \tString: Return the name of multipath device.
    """
    if libsan.host.linux.dist_name() == "RHEL" and libsan.host.linux.dist_ver() < 7:
        device = "/dev/%s" % disk
        mmnum_cmd = 'ls -l %s | awk \'{print "("$5,$6")"}\'' % device
        retcode, output = run(mmnum_cmd, return_output=True, verbose=False)
        if retcode != 0:
            _print("FAIL: Could not get device information for %s" % device)
            print(output)
            return None
        deps_cmd = rf"dmsetup deps | grep \"{output}\" | awk -v device={device} '{{print $1}}' | tr -d \"\:\""
        retcode, output = run(deps_cmd, return_output=True, verbose=False)
        if retcode != 0:
            _print("FAIL: Could not get dmsetup information for %s" % device)
            print(output)
            return None
        mpath = output
    else:
        # BZ891921 - Multipath provides the mpath device of given scsi block device
        cmd = "pidof multipathd"
        retcode, output = run(cmd, return_output=True, verbose=False)
        if retcode != 0:
            _print("FAIL: For some reason multipathd is not running")
            print(output)
            return None
        fmt = '"%d %m"'
        cmd = f"multipathd show paths format {fmt} | egrep \"^{disk}\" | awk '{{print$2}}'"
        retcode, output = run(cmd, return_output=True, verbose=False)
        if retcode != 0:
            _print("FAIL: to find multipath of disk %s" % disk)
            print(output)
            return None
        mpath = output

    if not mpath:
        _print("WARN: Could not find multipath for %s" % disk)
        return None
    return mpath


def get_disks_of_mpath(mpath_name):
    """Return all SCSI devices that belong to this mpath."""
    if not mpath_name:
        return None

    mpath_dict = multipath_query_all(mpath_name)
    if not mpath_dict:
        return None

    if "disk" not in list(mpath_dict.keys()):
        _print("WARN: mpath %s has no disk" % mpath_name)
        return None

    return list(mpath_dict["disk"].keys())


def get_disks_of_mpath_by_wwpn(mpath_name, wwpn):
    """From a specific mpath device, return the devices
    connected to this WWPN.
    """
    if not mpath_name or not wwpn:
        return None

    mpath_dict = multipath_query_all(mpath_name)
    if not mpath_dict:
        return None

    scsi_devices = []
    wwpn = libsan.host.fc.standardize_wwpn(wwpn)
    if wwpn and "disk" in list(mpath_dict.keys()):
        port_entries = ["h_wwpn", "t_wwpn"]
        for disk in list(mpath_dict["disk"].keys()):
            for port_entry in port_entries:
                if port_entry in list(mpath_dict["disk"][disk].keys()):  # noqa: SIM102
                    if mpath_dict["disk"][disk][port_entry] and wwpn in mpath_dict["disk"][disk][port_entry]:
                        scsi_devices.append(disk)
    return scsi_devices


def get_disks_of_mpath_by_iqn(mpath_name, iqn):
    """From a specific mpath device, return the devices
    connected to this IQN.
    """
    if not mpath_name or not iqn:
        return None

    if not libsan.host.iscsi.is_iqn(iqn):
        return None

    mpath_dict = multipath_query_all(mpath_name)
    if not mpath_dict:
        return None

    scsi_devices = []
    if "disk" in list(mpath_dict.keys()):
        port_entries = ["h_iqn", "t_iqn"]
        for disk in list(mpath_dict["disk"].keys()):
            for port_entry in port_entries:
                if port_entry in list(mpath_dict["disk"][disk].keys()):  # noqa: SIM102
                    if mpath_dict["disk"][disk][port_entry] and iqn in mpath_dict["disk"][disk][port_entry]:
                        scsi_devices.append(disk)
    return scsi_devices


def get_disks_of_mpath_by_mac(mpath_name, mac):
    """From a specific mpath device, return the devices
    connected to this MAC.
    """
    if not mpath_name or not mac:
        return None

    if not libsan.host.net.is_mac(mac):
        return None

    mpath_dict = multipath_query_all(mpath_name)
    if not mpath_dict:
        return None

    scsi_devices = []
    if "disk" in list(mpath_dict.keys()):
        port_entries = ["iface_mac"]
        for disk in list(mpath_dict["disk"].keys()):
            for port_entry in port_entries:
                if port_entry in list(mpath_dict["disk"][disk].keys()):  # noqa: SIM102
                    if mpath_dict["disk"][disk][port_entry] and mac in mpath_dict["disk"][disk][port_entry]:
                        scsi_devices.append(disk)
    return scsi_devices


def multipath_query_all(mpath_name=None):
    # Not sure with of these 2 commands I should use
    # multipath -ll
    # will force multipath to issue its own (synchronous) path checker call,
    # and report that result.
    # multipathd show top
    # multipath will not run a checker on the paths. It will simply report the
    # current state. This means that you get the results from the last time
    # that multipathd ran the path checker.  If there is no IO going to the
    # device, the kernel won't return an error on the path, and so you will
    # not see any errors until the next checker is run after you bring the
    # port down.
    cmd = "multipath -ll"
    if mpath_name:
        cmd += " %s" % mpath_name
    # cmd = "multipathd -k\"show top\""
    retcode, output = run(cmd, return_output=True, verbose=False)
    if retcode != 0:
        _print('FAIL: Could not execute "%s"' % cmd)
        return None

    if not output:
        # _print("FAIL: Got no output from \"%s\"" % cmd)
        return None

    regex_mpath = r"^(?:create\:\ ){0,1}"  # might have when ceating mpath
    regex_mpath += r"(?:reload\:\ ){0,1}"  # might have when ceating mpath
    regex_mpath += r"([^ ]+?)\ +"  # 1. mpath_name
    regex_mpath += r"(?:\((\S+)\)\ +){0,1}"  # 2. wwid if different from #1
    regex_mpath += r"(?:(dm\-[0-9]+)\ +){0,1}"  # 3. dm_name if known
    regex_mpath += r"(?:([^,]+),){0,1}"  # 4. vendor, might not have
    regex_mpath += r"(?:([^,]+)){0,1}$"  # 5. product,  might not have

    regex_feature = r"size=([0-9\.]+[A-Z])\ +"  # 1. size
    regex_feature += r"features=\'([^\']+)\'\ +"  # 2. features
    regex_feature += r"hwhandler=\'([^\']+)\'\ +"  # 3. hardware handler
    regex_feature += r"(?:wp=([rw]+)){0,1}"  # 4. write permission if known

    regex_pg = r"[^a-z]+policy=\'([^\`]+)\'\ +"  # 1. policy of this PG path_selector
    regex_pg += r"(?:prio=([0-9\-]+)\ +){0,1}"  # 2. priority of this PG if known
    regex_pg += r"(?:status=([a-z]+)){0,1}"  # 3. status of this PG if known (active or enable, etc)

    regex_disk = r"[^0-9]+"
    regex_disk += r"([0-9]+:[0-9]+:[0-9]+:[0-9]+)\ +"  # 1. scsi_id
    regex_disk += r"([a-z]+)\ +"  # 2. dev_name
    regex_disk += r"([0-9]+:[0-9]+)\ +"  # 3. major:minor
    regex_disk += r"(?:([a-z]+)\ +){0,1}"  # 4. dm_status if known, kernel level
    #   failed or active
    regex_disk += r"(?:([a-z]+)\ +){0,1}"  # 5. path_status if known, maintained
    #   by multipathd path_checker
    #   ready  or ghost or
    #   faulty or shaky
    regex_disk += r"(?:([a-z]+)){0,1}"  # 6. online_status: running or offline

    _mpath_name = None
    pg_id = None
    size = None
    size_human = None
    size_bytes = None
    wwid = None

    all_mpath_dict = {}
    all_mpath_dict["by_wwid"] = {}
    all_mpath_dict["by_mpath_name"] = {}
    all_mpath_dict["by_scsi_id"] = {}
    mpath_dict = {}

    scsi_host_id_2_driver = {}
    scsi_host_id_2_h_wwpn = {}
    fc_target_id_2_t_wwpn = {}

    all_scsi_host_ids = libsan.host.scsi.get_hosts()

    for host_id in all_scsi_host_ids:
        scsi_driver = libsan.host.scsi.scsi_driver_of_host_id(host_id)
        if scsi_driver:
            scsi_host_id_2_driver[host_id] = scsi_driver
        scsi_host_id_2_h_wwpn[host_id] = libsan.host.fc.get_fc_host_wwpn(host_id)
        t_wwpns_of_host = libsan.host.fc.t_wwpn_of_host(host_id)
        if t_wwpns_of_host:
            for t_wwpn in t_wwpns_of_host:
                t_ids = libsan.host.fc.fc_target_id_of_wwpn(t_wwpn)
                if t_ids:
                    for t_id in t_ids:
                        fc_target_id_2_t_wwpn[t_id] = t_wwpn

    lines = output.split("\n")

    for line in lines:
        m = re.match(regex_feature, line)
        if m:
            if not wwid or not mpath_dict:
                continue
            size = m.group(1)
            size_human = size + "iB"
            size_bytes = libsan.misc.size.size_human_2_size_bytes(size_human)
            mpath_dict["size_mp"] = size
            mpath_dict["size_human"] = size_human
            mpath_dict["size_bytes"] = size_bytes
            mpath_dict["feature"] = m.group(2)
            mpath_dict["hw_handleer"] = m.group(3)
            mpath_dict["permission"] = m.group(4)
            mpath_dict["disk"] = {}
            mpath_dict["t_wwpns"] = []
            mpath_dict["h_wwpns"] = []
            mpath_dict["t_iqns"] = []
            mpath_dict["h_iqns"] = []
            mpath_dict["map_info"] = []
            mpath_dict["iface_macs"] = []
            mpath_dict["iface_names"] = []
            mpath_dict["target_ips"] = []
            mpath_dict["persist_ips"] = []
            mpath_dict["transport_types"] = []
            mpath_dict["scsi_drivers"] = []
            mpath_dict["host_ids"] = []
            # _print("\n\nDEBUG - Feature - query all mpath [%s]" % line)
            # print all_mpath_dict["by_mpath_name"]

            continue

        m = re.match(regex_pg, line)
        if m:
            if not wwid or not mpath_dict:
                continue
            pg_id += 1
            mpath_dict["path_group"][pg_id] = {}
            mpath_dict["path_group"][pg_id]["wwid"] = wwid
            mpath_dict["path_group"][pg_id]["mpath_name"] = _mpath_name
            mpath_dict["path_group"][pg_id]["wwid"] = wwid
            mpath_dict["path_group"][pg_id]["pg_id"] = pg_id
            mpath_dict["path_group"][pg_id]["policy"] = m.group(1)
            mpath_dict["path_group"][pg_id]["prio"] = m.group(2)
            mpath_dict["path_group"][pg_id]["status"] = m.group(3)
            mpath_dict["path_group"][pg_id]["disk"] = {}
            # _print("\n\nDEBUG - PG : query all mpath [%s]" % line)
            # print all_mpath_dict["by_mpath_name"]

            continue

        m = re.match(regex_disk, line)
        if m:
            if not wwid or not mpath_dict or not pg_id:
                continue
            scsi_id = m.group(1)
            scsi_disk = m.group(2)
            major_minor = m.group(3)
            dm_status = m.group(4)
            path_status = m.group(5)
            online_status = m.group(6)
            # host ID is the first numbers of scsi_id
            host_id = None
            m = re.match(r"(\d+):.*", scsi_id)
            if m:
                host_id = m.group(1)

            t_wwpn = None
            h_wwpn = None
            h_iqn = None
            t_iqn = None
            iface_name = None
            iface_mac = None
            target_ip = None
            persist_ip = None
            scsi_driver = "N/A"

            match_scsi_id = re.search(libsan.host.scsi.get_regex_scsi_id(), line)
            if match_scsi_id:
                scsi_host_id = match_scsi_id.group(1)
                fc_target_id = f"{match_scsi_id.group(1)}:{match_scsi_id.group(2)}:{match_scsi_id.group(3)}"
                if scsi_host_id in list(scsi_host_id_2_driver.keys()):
                    scsi_driver = scsi_host_id_2_driver[scsi_host_id]

                if scsi_host_id in list(scsi_host_id_2_h_wwpn.keys()):
                    h_wwpn = scsi_host_id_2_h_wwpn[scsi_host_id]

                if h_wwpn:
                    if fc_target_id in list(fc_target_id_2_t_wwpn.keys()):
                        t_wwpn = fc_target_id_2_t_wwpn[fc_target_id]
                else:
                    iscsi_session = libsan.host.iscsi.get_iscsi_session_by_scsi_id(scsi_id)
                    if iscsi_session:
                        h_iqn = iscsi_session["h_iqn"]
                        t_iqn = iscsi_session["t_iqn"]
                        iface_name = iscsi_session["iface"]
                        iface_mac = iscsi_session["mac"]
                        target_ip = iscsi_session["target_ip"]
                        persist_ip = iscsi_session["persist_ip"]

            disk_info_dict = {}
            disk_info_dict["scsi_id"] = scsi_id
            disk_info_dict["host_id"] = host_id
            disk_info_dict["scsi_disk"] = scsi_disk
            disk_info_dict["scsi_driver"] = scsi_driver
            disk_info_dict["pg_id"] = pg_id
            disk_info_dict["mpath_name"] = _mpath_name
            disk_info_dict["wwid"] = wwid
            disk_info_dict["dm_status"] = dm_status
            disk_info_dict["path_status"] = path_status
            disk_info_dict["online_status"] = online_status
            disk_info_dict["major_minor"] = major_minor
            disk_info_dict["size"] = size
            disk_info_dict["size_human"] = size_human
            disk_info_dict["size_bytes"] = size_bytes
            disk_info_dict["t_wwpn"] = t_wwpn
            disk_info_dict["h_wwpn"] = h_wwpn
            disk_info_dict["t_iqn"] = t_iqn
            disk_info_dict["h_iqn"] = h_iqn
            disk_info_dict["iface_name"] = iface_name
            disk_info_dict["iface_mac"] = iface_mac
            disk_info_dict["target_ip"] = target_ip
            disk_info_dict["persist_ip"] = persist_ip
            disk_info_dict["transport_type"] = "UNKNOWN"

            if h_iqn:
                disk_info_dict["transport_type"] = "iSCSI"

            if h_wwpn:
                disk_info_dict["transport_type"] = libsan.host.fc.fc_host_transport_type(
                    libsan.host.fc.fc_host_id_of_wwpn(h_wwpn)
                )

            if scsi_driver == "scsi_debug":
                disk_info_dict["transport_type"] = "scsi_debug"

            # Each disk has its own information, we should add each disk from mpath info to mpath info
            if disk_info_dict["transport_type"] not in mpath_dict["transport_types"]:
                mpath_dict["transport_types"].append(disk_info_dict["transport_type"])

            if scsi_driver and scsi_driver not in mpath_dict["scsi_drivers"]:
                mpath_dict["scsi_drivers"].append(scsi_driver)

            if host_id and host_id not in mpath_dict["host_ids"]:
                mpath_dict["host_ids"].append(host_id)

            if t_wwpn and t_wwpn not in mpath_dict["t_wwpns"]:
                mpath_dict["t_wwpns"].append(t_wwpn)

            if h_wwpn and h_wwpn not in mpath_dict["h_wwpns"]:
                mpath_dict["h_wwpns"].append(h_wwpn)

            if h_wwpn and t_wwpn:
                map_info = {"t_wwpn": t_wwpn, "h_wwpn": h_wwpn}
                mpath_dict["map_info"].append(map_info)

            if t_iqn and t_iqn not in mpath_dict["t_iqns"]:
                mpath_dict["t_iqns"].append(t_iqn)

            if h_iqn and h_iqn not in mpath_dict["h_iqns"]:
                mpath_dict["h_iqns"].append(h_iqn)

            if h_iqn and t_iqn:
                map_info = {"t_iqn": t_iqn, "h_iqn": h_iqn}
                mpath_dict["map_info"].append(map_info)

            if iface_mac and iface_mac not in mpath_dict["iface_macs"]:
                mpath_dict["iface_macs"].append(iface_mac)

            if iface_name and h_iqn not in mpath_dict["iface_names"]:
                mpath_dict["iface_names"].append(iface_name)

            if target_ip and target_ip not in mpath_dict["target_ips"]:
                mpath_dict["target_ips"].append(target_ip)

            if persist_ip and persist_ip not in mpath_dict["persist_ips"]:
                mpath_dict["persist_ips"].append(persist_ip)

            mpath_dict["disk"][disk_info_dict["scsi_disk"]] = disk_info_dict
            mpath_dict["path_group"][pg_id]["disk"][scsi_id] = disk_info_dict
            all_mpath_dict["by_scsi_id"][scsi_id] = disk_info_dict
            # _print("\n\nDEBUG - Disk: query all mpath [%s]" % line)
            # print all_mpath_dict["by_mpath_name"]

            continue

        # as regex_mpath is an agressive regex, we leave it at the last
        # one to check.
        m = re.match(regex_mpath, line)
        if m:
            if re.match(r"^\|", line):
                continue

            _mpath_name = m.group(1)
            pg_id = 0
            wwid = _mpath_name
            if m.group(2):
                wwid = m.group(2)

            dm_name = None
            if m.group(3):
                dm_name = m.group(3)

            vendor_raw = None
            vendor = vendor_raw
            if m.group(4):
                vendor_raw = m.group(4)
                vendor = re.sub("[ ]+$", "", vendor_raw)

            product_raw = None
            product = product_raw
            if m.group(5):
                product_raw = m.group(5)
                product = re.sub("[ ]+$", "", product_raw)

            mpath_dict = {}
            mpath_dict["mpath_name"] = _mpath_name
            mpath_dict["wwid"] = wwid
            mpath_dict["dm_name"] = dm_name
            mpath_dict["vendor"] = vendor
            mpath_dict["vendor_raw"] = vendor_raw
            mpath_dict["product"] = product
            mpath_dict["product_raw"] = product_raw
            mpath_dict["h_wwpns"] = []
            mpath_dict["t_wwpns"] = []
            mpath_dict["disk"] = {}
            mpath_dict["path_group"] = {}
            mpath_dict["transport_types"] = []

            all_mpath_dict["by_wwid"][wwid] = mpath_dict
            all_mpath_dict["by_mpath_name"][_mpath_name] = mpath_dict
            # _print("\n\nDEBUG: query all mpath [%s]" % line)
            # print all_mpath_dict["by_mpath_name"]
            continue
    if mpath_name:
        if mpath_name in list(all_mpath_dict["by_mpath_name"].keys()):
            return all_mpath_dict["by_mpath_name"][mpath_name]
        return None

    return all_mpath_dict


def mpath_name_of_wwid(wwid):
    if not wwid:
        _print("FAIL: mpath_name_of_wwid() - requires wwid parameter")
        return None

    mp_devs = multipath_query_all()
    if not mp_devs:
        return None

    if wwid in list(mp_devs["by_wwid"].keys()):
        return mp_devs["by_wwid"][wwid]["mpath_name"]

    return None


def mpath_names_of_vendor(vendor):
    """Given a Vendor return a list with all multipath device names with this vendor."""
    if not vendor:
        _print("FAIL: mpath_names_of_vendor() - requires vendor parameter")
        return None

    mp_devs = multipath_query_all()
    if not mp_devs:
        return None

    mpaths = []
    for mpath_info in list(mp_devs["by_mpath_name"].values()):
        if "vendor" in list(mpath_info.keys()) and mpath_info["vendor"] == vendor:
            mpaths.append(mpath_info["mpath_name"])

    return mpaths


def mpath_check_disk_status(mpath_name, scsi_disk):
    """Check the online status of a specific SCSI disk from a mpath device."""
    if not mpath_name or not scsi_disk:
        _print("FAIL: mpath_check_disk_status() - requires mpath_name and scsi_disk parameters")
        return None

    mpath_dict = multipath_query_all(mpath_name)
    if not mpath_dict:
        _print("FAIL: mpath_check_disk_status() - Could not get mpath info for %s" % mpath_name)
        return None

    if "disk" not in list(mpath_dict.keys()):
        _print("FAIL: mpath_check_disk_status() - Could not find any SCSI disk for %s" % mpath_name)
        return None

    for disk in list(mpath_dict["disk"].keys()):
        if scsi_disk == mpath_dict["disk"][disk]["scsi_disk"]:
            return mpath_dict["disk"][disk]["online_status"]

    return None


def mpath_check_disk_dm_status(mpath_name, scsi_disk):
    """Check the dm status of a specific SCSI disk from a mpath device."""
    if not mpath_name or not scsi_disk:
        _print("FAIL: mpath_check_disk_dm_status() - requires mpath_name and scsi_disk parameters")
        return None

    mpath_dict = multipath_query_all(mpath_name)
    if not mpath_dict:
        _print("FAIL: mpath_check_disk_dm_status() - Could not get mpath info for %s" % mpath_name)
        return None

    if "disk" not in list(mpath_dict.keys()):
        _print("FAIL: mpath_check_disk_dm_status() - Could not find any SCSI disk for %s" % mpath_name)
        return None

    for disk in list(mpath_dict["disk"].keys()):
        if scsi_disk == mpath_dict["disk"][disk]["scsi_disk"]:
            return mpath_dict["disk"][disk]["dm_status"]

    return None


def mpath_check_disk_path_status(mpath_name, scsi_disk):
    """Check the path status of a specific SCSI disk from a mpath device."""
    if not mpath_name or not scsi_disk:
        _print("FAIL: mpath_check_disk_path_status() - requires mpath_name and scsi_disk parameters")
        return None

    mpath_dict = multipath_query_all(mpath_name)
    if not mpath_dict:
        _print("FAIL: mpath_check_disk_path_status() - Could not get mpath info for %s" % mpath_name)
        return None

    if "disk" not in list(mpath_dict.keys()):
        _print("FAIL: mpath_check_disk_path_status() - Could not find any SCSI disk for %s" % mpath_name)
        return None

    for disk in list(mpath_dict["disk"].keys()):
        if scsi_disk == mpath_dict["disk"][disk]["scsi_disk"]:
            return mpath_dict["disk"][disk]["path_status"]

    return None


def mpath_get_active_disk(mpath_name):
    """From specific mpath device get which disk is the active one
    Return
    \tList of active SCSI disks.
    """
    if not mpath_name:
        _print("FAIL: mpath_get_active_disk() - requires mpath_name parameter")
        return None

    mpath_dict = multipath_query_all(mpath_name)
    if not mpath_dict:
        _print("FAIL: mpath_get_active_disk() - Could not get mpath info for %s" % mpath_name)
        return None

    if "disk" not in list(mpath_dict.keys()):
        _print("FAIL: mpath_get_active_disk() - Could not find any SCSI disk for %s" % mpath_name)
        return None

    active_disks = []
    for disk in list(mpath_dict["disk"].keys()):
        pg = mpath_dict["disk"][disk]["pg_id"]
        if mpath_dict["path_group"][pg]["status"] == "active":
            active_disks.append(disk)

    if active_disks:
        return active_disks

    _print("FAIL: mpath_get_active_disk() - Could not find any active disk for %s" % mpath_name)
    return None


def get_free_mpaths(exclude_boot_device=True, exclude_lvm_device=True):
    """Return a dict of free mpath devices."""
    all_mp_info = multipath_query_all()
    if not all_mp_info:
        # could not query multipath devices
        return None

    if "by_wwid" not in list(all_mp_info.keys()):
        # mpath device was not found
        print(list(all_mp_info.keys()))
        return None

    pvs = libsan.host.lvm.pv_query()
    boot_dev = libsan.host.linux.get_boot_device()
    boot_wwid = None
    # if for some reason we boot from a single disk, but this disk is part of multipath device
    # the mpath device should be skipped as well
    if boot_dev:
        boot_wwid = libsan.host.linux.get_device_wwid(boot_dev)

    choosed_mpaths = {}

    for mp_wwid in list(all_mp_info["by_wwid"].keys()):
        mp_info = all_mp_info["by_wwid"][mp_wwid]
        # Skip if mpath device is used for boot
        if boot_wwid == mp_info["wwid"] and exclude_boot_device:
            _print("DEBUG: get_free_mpaths() - skip %s as it is used for boot" % mp_info["mpath_name"])
            continue

        # Skip if it is used by LVM
        if pvs and exclude_lvm_device:
            mp_used_by_lvm = False
            for pv in list(pvs.keys()):
                if mpath_device_2_mpath_name(pv) == mp_info["mpath_name"]:
                    mp_used_by_lvm = True
                    _print("DEBUG: get_free_mpaths() - skip %s as it is used for LVM" % mp_info["mpath_name"])
                    continue
            if mp_used_by_lvm:
                continue

        choosed_mpaths[mp_info["mpath_name"]] = mp_info

    return choosed_mpaths


def h_wwpns_of_mpath(mpath_name):
    """Return the h_wwpns of specific mpath device."""
    if not mpath_name:
        _print("FAIL: h_wwpns_of_mpath() - requires mpath_name parameter")
        return None

    mpath_dict = multipath_query_all(mpath_name)
    if not mpath_dict:
        _print("FAIL: h_wwpns_of_mpath() - Could not get mpath info for %s" % mpath_name)
        return None

    if "h_wwpns" in mpath_dict:
        return mpath_dict["h_wwpns"]
    return None


def t_wwpns_of_mpath(mpath_name):
    """Return the h_wwpns of specific mpath device."""
    if not mpath_name:
        _print("FAIL: t_wwpns_of_mpath() - requires mpath_name parameter")
        return None

    mpath_dict = multipath_query_all(mpath_name)
    if not mpath_dict:
        _print("FAIL: t_wwpns_of_mpath() - Could not get mpath info for %s" % mpath_name)
        return None

    if "t_wwpns" in mpath_dict:
        return mpath_dict["t_wwpns"]
    return None


def iface_macs_of_mpath(mpath_name):
    """Return the iface_macs_of_mpath of specific mpath device."""
    if not mpath_name:
        _print("FAIL: iface_macs_of_mpath() - requires mpath_name parameter")
        return None

    mpath_dict = multipath_query_all(mpath_name)
    if not mpath_dict:
        _print("FAIL: iface_macs_of_mpath() - Could not get mpath info for %s" % mpath_name)
        return None

    if "iface_macs" in mpath_dict:
        return mpath_dict["iface_macs"]
    return None


def transport_types_of_mpath(mpath_name):
    """Return the transport_types of specific mpath device."""
    if not mpath_name:
        _print("FAIL: transport_types_of_mpath() - requires mpath_name parameter")
        return None

    mpath_dict = multipath_query_all(mpath_name)
    if not mpath_dict:
        _print("FAIL: transport_types_of_mpath() - Could not get mpath info for %s" % mpath_name)
        return None

    if "transport_types" in mpath_dict:
        return mpath_dict["transport_types"]
    return None


def multipath_show(mpath_name=None):
    cmd = "multipath -ll"
    if mpath_name:
        cmd += " %s" % mpath_name
    run(cmd)


def multipath_reload(mpath_name=None):
    """Usage
    multipath_reload()
    Purpose
        Execute 'multipath -r'
    Parameter
        N/A
    Returns
        1
            or
        undef
    Exceptions
        N/A.
    """
    cmd = "multipath -r"
    if mpath_name:
        cmd += " %s" % mpath_name
    ret = run(cmd, verbose=True)
    if ret != 0:
        return False

    return True


def remove_mpath(mpath_name):
    """Remove specific mpath."""
    if not mpath_name:
        _print("FAIL: remove_mpath() - requires mpath_name parameter")
        return False
    retcode = run("multipath -f %s" % mpath_name)
    if retcode != 0:
        _print("FAIL: Could not remove mpath %s" % mpath_name)
        return False
    return True


def flush_all():
    """Flush all unused multipath device maps."""
    cmd = "multipath -F"
    ret = run(cmd, verbose=True)
    if ret != 0:
        return False

    return True


def multipath_backup_conf(bak_file):
    """backup_mp_conf ()
    Usage
    backup_mp_conf(bak_file)
    Purpose
    Check if bak_file exists, if not, copy current to it.
    Parameter
    bak_file       # like "/etc/multipath.conf.bak"
    Returns
    true
        or
    False           # file exists or source file not exists;
    Exceptions
    N/A.
    """

    if not os.path.isfile(multipath_conf_path):
        _print("FAIL: %s does not exist" % multipath_conf_path)
        return False

    if os.path.isfile(bak_file):
        _print("FAIL: mpath backup file %s already exists" % bak_file)
        return False

    _print(f"INFO: Backing up {multipath_conf_path} to {bak_file}")
    cmd = f"cp -f {multipath_conf_path} {bak_file}"
    ret, output = run(cmd, return_output=True)
    if ret != 0:
        _print("FAIL: Could not backup %s" % multipath_conf_path)
        print(output)
        return False

    return True


def multipath_restore_conf(bak_file):
    """multipath_restore_conf ()
    Usage
    multipath_restore_conf(bak_file)
    Purpose
    Check if bak_file exists, if so, copy it to '/etc/multipath.conf'
    and reload mpath conf.
    Parameter
    bak_file       # like "/etc/multipath.conf.bak"
    Returns
    True
        or
    False           # file exists or source file not exists;
    Exceptions
    N/A.
    """
    if not os.path.isfile(bak_file):
        _print("FAIL: %s does not exist" % bak_file)
        return False

    _print("INFO: Restoring multipath configuration from %s" % bak_file)
    cmd = f"cp -f {bak_file} {multipath_conf_path}"
    ret, output = run(cmd, return_output=True)
    if ret != 0:
        _print("FAIL: Could not restore backup from %s" % bak_file)
        print(output)
        return False

    multipath_reload_conf()
    return True


def multipath_reload_conf():
    """Usage
    multipath_reload_conf()
    Purpose
        Execute "multiapthd -k'reconfig'" to load configuration.
        After that execute 'multipath -r'
    Parameter
        N/A
    Returns
        1
            or
        undef
    Exceptions
        N/A.
    """
    cmd = "multipathd -k'reconfig'"
    ret = run(cmd, verbose=True)

    multipath_reload()
    if ret != 0:
        return False

    return True


def multipath_setup_blacklist():
    """multipath_setup_blacklist ()
    Usage
    multipath_setup_blacklist()
    Purpose
    Add 'wwid .*' into blacklist and all exists mpath wwid to
    'blacklist_exceptions'. This will prevent mpath take over newly
    created LUN. This function DO NOT back up configuration, use
    backup_mp_conf() in stead.
    Parameter
    N/A
    Returns
    1
        or
    undef
    Exceptions
    N/A.
    """
    all_mp_info_dict = multipath_query_all()
    if not all_mp_info_dict:
        return False
    current_wwids = list(all_mp_info_dict["by_wwid"].keys())
    if not current_wwids:
        return False

    if not mpath_conf_set("/blacklist/wwid", ".*"):
        return False

    for wwid in current_wwids:
        if not mpath_conf_set("/blacklist_exceptions/wwid[last()+1]", wwid):
            return False

    _print("INFO: Reloading updated multipath configuration")
    multipath_reload_conf()
    return True


def mpath_conf_remove(path, value):
    """Remove parameter from multipath config using augeas.
    For non-unique paths, use return of mptah_conf_match as path
    For valid options use 'man multipath.conf'.

    :param path: eg. "/blacklist/wwid"
    :param value: eg. "<device wwid>"
    :returns: boolean
    :raises IOError: Augeas save fails. Check if multipath.conf changes are valid.
    """
    matched_path = mpath_conf_match(path, value)

    if matched_path:
        _, out = run(f'augtool -s rm "{matched_path}"', return_output=True, verbose=False)
        if "Saved 1 file(s)" not in out:
            _print(f"FAIL: unable to set {path} to {value}")
            return False

    return True


def mpath_conf_match(path, value):
    """Checks if parameter is set in multipath.conf using augeas
    Use path expressions to get non-unique paths: https://github.com/hercules-team/augeas/wiki/Path-expressions
    e.g. path="/devices/device[*]/vendor" value="LIO-ORG"
    Successfully matches even if "end of the path" is not unique. e.g. path="/blacklist_exceptions/wwid" value="$wwid".

    :param path: eg. "/blacklist/wwid"
    :param value: eg. "<device wwid>"
    :returns: matched path or None
    """
    if not libsan.host.linux.install_package("augeas", verbose=False):
        _print("FAIL: Could not install augeas")
        return None

    full_path = path if path.startswith(aug_conf_path) else aug_conf_path + path

    # successfully matches also when end of path is not unique - path[*] value, but not path[*]/path value
    _, out = run(f'augtool match "{full_path}" "{value}"', return_output=True, verbose=False)

    # augtool prints path when matched, return codes seems useless
    if aug_conf_path in out:
        # returning path, as it can be useful when modifying non-unique paths
        return out
    return None


def mpath_conf_set(path, value):
    """Change multipath.conf parameters using augeas.
    To append to non-unique paths, use [last()+1].
    eg. path="/devices/device[last()+1]/vendor" value="LIO-ORG"
    For valid options use 'man multipath.conf'.

    :param path: eg. "/defaults/path_selector" or "/files/etc/multipath.conf/defaults/path_selector"
    :param value: eg. "round-robin"
    :returns: boolean
    :raises IOError: Augeas save fails. Check if multipath.conf changes are valid.
    """
    full_path = path if path.startswith(aug_conf_path) else aug_conf_path + path

    # Need to check if not already exists to avoid duplicates
    path_to_match = re.sub(r"\[[^]]*]", "[.]", path)  # in case path expressions are being used
    matched_path = mpath_conf_match(path_to_match, value)

    if not matched_path:
        if not libsan.host.linux.is_installed("augeas"):
            _print("FAIL: Could not install augeas")
            return False
        _, out = run(f'augtool -s set "{full_path}" "{value}"', return_output=True, verbose=False)
        if "Saved 1 file(s)" not in out:
            _print(f"FAIL: unable to set {path} to {value}. Try")
            return False

    return True


def mp_query_conf(config_str):
    """Parse string containing multipath config to a dict."""
    regex_option = r"^[ \t]*"
    regex_option += r"([^\ #=\t]+)"
    regex_option += r"[\ \t]+"
    regex_option += r"(?:'|\"){0,1}"
    regex_option += r"([^'\"]+)"
    regex_option += r"(?:'|\"){0,1}"
    regex_option += r"(?:\#.*){0,1}"  # we remove '|"

    regex_section = r"[ \t]*([a-z_]+)[ \t]*\{"
    if not config_str:
        return None

    config_dict = {}

    current_section = None
    section_name = None
    for line in config_str.split("\n"):
        m = re.match(regex_section, line)
        if m:
            key = m.group(1)
            if key == "device":
                if not section_name:
                    continue
                if "devs" not in list(config_dict[section_name].keys()):
                    config_dict[section_name]["devs"] = []
                # We are in a subsection, need to update current_section
                tmp_dict = {}
                config_dict[section_name]["devs"].append(tmp_dict)
                # poiting current_section to last item in the list
                current_section = config_dict[section_name]["devs"][-1]
                continue

            if key == "multipath":
                if not section_name:
                    continue
                if "mpaths" not in list(config_dict[section_name].keys()):
                    config_dict[section_name]["mpaths"] = []
                # We are in a subssection, need to update current_section
                tmp_dict = {}
                config_dict[section_name]["mpaths"].append(tmp_dict)
                # poiting current_section to last item in the list
                current_section = config_dict[section_name]["mpaths"][-1]
                continue

            if key not in config_dict:
                config_dict[key] = {}
            current_section = config_dict[key]
            section_name = key
            continue
        m = re.match(regex_option, line)
        if m:
            key = m.group(1)
            value = m.group(2)
            if current_section is None:
                continue
            if key == "devnode":
                if "devnodes" not in list(current_section.keys()):
                    current_section["devnodes"] = []
                current_section["devnodes"].append(value)
                continue

            match_section = re.match("^blacklist", section_name)
            if key == "wwid" and match_section:
                if "wwids" not in list(current_section.keys()):
                    current_section["wwids"] = []
                current_section["wwids"].append(value)
                continue

            current_section[key] = value
            continue

    return config_dict


def mp_query_saved_conf():
    """Usage
        mp_query_saved_conf()
    Purpose
        Load multipath config file and return it as a dict
    Parameter
        N/A
    Returns
        mp_conf_dict.
    """
    if not os.path.isfile(multipath_conf_path):
        _print("FAIL: %s does not exist" % multipath_conf_path)
        return None

    cfg_text = _load_config(multipath_conf_path)
    if not cfg_text:
        return None

    return mp_query_conf(cfg_text)


def _load_config(config_file):
    """Parse multipath config file to dict."""
    with open(config_file) as f:
        return f.read()


def _save_config(config_dict, config_file=multipath_conf_path):
    """Convert multipath config dict to string and save to file."""
    config_str = ""
    if "defaults" in list(config_dict.keys()):
        config_str += "defaults {\n"
        for key in list(config_dict["defaults"].keys()):
            config_str += "\t{} {}\n".format(key, config_dict["defaults"][key])
        config_str += "}\n"

    blacklist_sections = ["blacklist", "blacklist_exceptions"]
    for b_section in blacklist_sections:
        if b_section in list(config_dict.keys()):
            config_str += "%s {\n" % b_section

            if "devnodes" in list(config_dict[b_section].keys()):
                for node in config_dict[b_section]["devnodes"]:
                    config_str += '\tdevnode "%s"\n' % node

            if "wwids" in list(config_dict[b_section].keys()):
                for wwid in config_dict[b_section]["wwids"]:
                    config_str += '\twwid "%s"\n' % wwid

            if "devs" in list(config_dict[b_section].keys()):
                for device in config_dict[b_section]["devs"]:
                    config_str += "\tdevice {\n"
                    for key in list(device.keys()):
                        config_str += f"\t\t{key} {device[key]}\n"
                    config_str += "\t}\n"

            config_str += "}\n"

    if "devices" in list(config_dict.keys()):
        config_str += "devices {\n"
        for vendor in list(config_dict["devices"].keys()):
            for product in config_dict["devices"][vendor]:
                config_str += "\tdevice {\n"
                for name in list(product.keys()):
                    config_str += f"\t\t{name} {product[name]}\n"
                config_str += "\t}\n"
        config_str += "}\n"

    if "multipaths" in list(config_dict.keys()):
        config_str += "multipaths {\n"
        for key in list(config_dict["multipaths"].keys()):
            config_str += "\tmultipath {\n"
            for wwid in config_dict["multipaths"][key]:
                config_str += f"\t\t{key} {wwid}\n"
            config_str += "}\n"
        config_str += "}\n"

    with open(config_file, "w") as f:
        f.write(config_str)

    return True


def mpath_device_2_mpath_name(mpath_device):
    """Convert a specific multipath device to mpath_name
    E.g. /dev/mapper/mpathap1 => mpatha.
    """
    multipath_dev_regex = r"/dev/mapper\/(.*)"
    m = re.match(multipath_dev_regex, mpath_device)
    if m:
        # need to remove partition information
        # Multipath names has changed on RHEL7 and if a device name ends in letter the
        # partition number is just append to the device number, if it ends in digit
        # it appends 'p' before the partition number
        # So we can have device partitions as below...
        # device_name = "mapper/360a98000572d5765636f69746f6a4f6a1"
        # device_name = "mapper/360a98000572d5765636f69746f6a4f61p1"
        device_name = m.group(1)
        m = re.match(r"(.*)p?\d", device_name)
        if not m:
            # does not seem to be a valid mpath device
            return None
        device_name = m.group(1)
        # remove trailing p if it exists
        device_name = re.sub(r"(\S+)p$", r"\1", device_name)
        return device_name
    return None



# Copyright (C) 2016 Red Hat, Inc.
# This file is part of libsan.
#
# libsan is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# libsan is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with libsan.  If not, see <http://www.gnu.org/licenses/>.

"""lsm.py: Module to manipulate libstoragemgmt userspace package."""

__author__ = "Jakub Krysl"
__copyright__ = "Copyright (c) 2017 Red Hat, Inc. All rights reserved."

import os
from time import sleep

import libsan.host.linux
from libsan import _print
from libsan.host.cli_tools import (
    FailedCheckExceptionError,
    Wrapper,
    WrongArgumentExceptionError,
    WrongCommandExceptionError,
)
from libsan.host.cmdline import run


def _cli(func):
    # This is a decorator to mark functions callable by 'lsmcli'
    func.cli = True
    return func


class LibStorageMgmt(Wrapper):
    def __init__(self, username=None, password=None, target=None, protocol=None, disable_check=False):
        self.disable_check = disable_check
        if libsan.host.linux.dist_ver() < 7:
            _print("FATAL: libstoragemgmt is not supported on RHEL < 7.")

        self.username = None
        self.password = None
        self.target = None
        self.protocol = None
        self.port = None
        self.query_params = None
        self.timeout = None

        # persistent previous values
        self.previous_sys_read_pct = {}
        self.previous_phy_disk_cache_policy = {}
        self.previous_read_cache_policy = {}
        self.previous_write_cache_policy = {}
        self.previous_local_disk_ident_led = {}
        self.previous_local_disk_fault_led = {}

        # local target does not require anything of this and megaraid/sim needs only protocol
        if username and password and target and protocol:
            self.username = username
            self.password = password
            self.target = target
            self.protocol = protocol
        elif protocol and "megaraid" in protocol or "sim" in protocol:
            self.protocol = protocol

        if self.password:
            os.environ["LSMCLI_PASSWORD"] = self.password
            _print("INFO: Password set")
        else:
            if os.environ.get("LSMCLI_PASSWORD"):
                del os.environ["LSMCLI_PASSWORD"]
                _print("INFO: Password cleaned")

        requires_restart = False
        # stop if can't install lsm package
        if not libsan.host.linux.is_installed("libstoragemgmt"):
            if not libsan.host.linux.install_package("libstoragemgmt", check=False):
                _print("FATAL: Could not install libstoragemgmt package")
            else:
                requires_restart = True

        if self.protocol == "smispy":
            self.port = "5988"
            self.query_params = "?namespace=root/emc"
        if self.protocol == "smispy+ssl":
            # ssl uses different port
            self.port = "5989"
            # ignore missing ssl certificate
            self.query_params = "?namespace=root/emc&no_ssl_verify=yes"

        # install protocol specific packages
        if self.protocol:
            if "ontap" in self.protocol:
                if not libsan.host.linux.is_installed("libstoragemgmt-netapp-plugin"):
                    if not libsan.host.linux.install_package("libstoragemgmt-netapp-plugin", check=False):
                        _print("FATAL: Could not install LSM NetApp plugin")
                    else:
                        requires_restart = True
            elif "smispy" in self.protocol:
                if not libsan.host.linux.is_installed("libstoragemgmt-smis-plugin"):
                    if not libsan.host.linux.install_package("libstoragemgmt-smis-plugin", check=False):
                        _print("FATAL: Could not install LSM SMIS plugin")
                    else:
                        requires_restart = True
            elif "targetd" in self.protocol:
                if not libsan.host.linux.is_installed("libstoragemgmt-targetd-plugin"):
                    if not libsan.host.linux.install_package("libstoragemgmt-targetd-plugin", check=False):
                        _print("FATAL: Could not install LSM targetd plugin")
                    else:
                        requires_restart = True
            elif "megaraid" in self.protocol:
                if not libsan.host.linux.is_installed("libstoragemgmt-megaraid-plugin"):
                    if not libsan.host.linux.install_package("libstoragemgmt-megaraid-plugin", check=False):
                        _print("FATAL: Could not install LSM megaraid plugin")
                    else:
                        requires_restart = True
                # needs to install 3rd party tool
                if not libsan.host.linux.install_package("storcli"):
                    _print("FATAL: Could not install storcli")

        if requires_restart:
            if run("service libstoragemgmt restart", verbose=True) != 0:
                _print("FATAL: Could not restart libstoragemgmt service")
            else:
                _print("INFO: Waiting for service to restart.")
                sleep(5)
        else:
            # check if libstoragemgmt service is running
            if libsan.host.linux.service_status("libstoragemgmt") != 0:  # noqa: SIM102
                if libsan.host.linux.service_start("libstoragemgmt"):
                    _print("INFO: Waiting for service to start.")
                    sleep(5)

        self.commands = {
            "list": "list",
            "job_status": "job-status",
            "capabilities": "capabilities",
            "plugin_info": "plugin-info",
            "volume_create": "volume-create",
            "volume_raid_create": "volume-raid-create",
            "volume_raid_create_cap": "volume-raid-create-cap",
            "volume_delete": "volume-delete",
            "volume_resize": "volume-resize",
            "volume_replicate": "volume-replicate",
            "volume_replicate_range": "volume-replicate-range",
            "volume_replicate_range_block_size": "volume-replicate-range-block-size",
            "volume_dependants": "volume-dependants",
            "volume_dependants_rm": "volume-dependants-rm",
            "volume_access_group": "volume-access-group",
            "volume_mask": "volume-mask",
            "volume_unmask": "volume-unmask",
            "volume_enable": "volume-enable",
            "volume_disable": "volume-disable",
            "volume_raid_info": "volume-raid-info",
            "volume_ident_led_on": "volume-ident-led-on",
            "volume_ident_led_off": "volume-ident-led-off",
            "system_read_cache_pct_update": "system-read-cache-pct-update",
            "pool_member_info": "pool-member-info",
            "access_group_create": "access-group-create",
            "access_group_add": "access-group-add",
            "access_group_remove": "access-group-remove",
            "access_group_delete": "access-group-delete",
            "access_group_volumes": "access-group-volumes",
            "iscsi_chap": "iscsi-chap",
            "fs_create": "fs-create",
            "fs_delete": "fs-delete",
            "fs_resize": "fs-resize",
            "fs_export": "fs-export",
            "fs_unexport": "fs-unexport",
            "fs_clone": "fs-clone",
            "fs_snap_create": "fs-snap-create",
            "fs_snap_delete": "fs-snap-delete",
            "fs_snap_restore": "fs-snap-restore",
            "fs_dependants": "fs-dependants",
            "fs_dependants_rm": "fs-dependants-rm",
            "file_clone": "file-clone",
            "local_disk_list": "local-disk-list",
            "volume_cache_info": "volume-cache-info",
            "volume_phy_disk_cache_update": "volume-phy-disk-cache-update",
            "volume_read_cache_policy_update": "volume-read-cache-policy-update",
            "volume_write_cache_policy_update": "volume-write-cache-policy-update",
            "local_disk_ident_led_on": "local-disk-ident-led-on",
            "local_disk_ident_led_off": "local-disk-ident-led-off",
            "local_disk_fault_led_on": "local-disk-fault-led-on",
            "local_disk_fault_led_off": "local-disk-fault-led-off",
        }
        self.commands["all"] = list(self.commands.keys())

        self.arguments = {
            "human": [self.commands["all"], " --human"],
            "terse": [self.commands["all"], " --terse="],
            "enum": [self.commands["all"], " --enum"],
            "force": [self.commands["all"], " --force"],
            "wait": [self.commands["all"], " --wait="],
            "header": [self.commands["all"], " --header"],
            "async": [self.commands["all"], " --b"],
            "script": [self.commands["all"], " --script"],
            "lsm_type": [["list"], " --type="],
            "sys": [
                [
                    "list",
                    "capabilities",
                    "volume_raid_create_cap",
                    "volume_replicate_range_block_size",
                    "system_read_cache_pct_update",
                    "access_group_create",
                ],
                " --sys=",
            ],
            "pool": [["list", "volume_create", "volume_replicate", "pool_member_info", "fs_create"], " --pool="],
            "vol": [
                [
                    "list",
                    "volume_delete",
                    "volume_resize",
                    "volume_replicate",
                    "volume_dependants",
                    "volume_dependants_rm",
                    "volume_access_group",
                    "volume_mask",
                    "volume_unmask",
                    "volume_enable",
                    "volume_disable",
                    "volume_raid_info",
                    "volume_ident_led_on",
                    "volume_ident_led_off",
                    "volume_cache_info",
                    "volume_phy_disk_cache_update",
                    "volume_read_cache_policy_update",
                    "volume_read_cache_policy_update",
                ],
                " --vol=",
            ],
            "disk": [["list", "volume_raid_create"], " --disk="],
            "ag": [
                [
                    "list",
                    "volume_mask",
                    "volume_unmask",
                    "access_group_add",
                    "access_group_remove",
                    "access_group_delete",
                    "access_group_volumes",
                ],
                " --ag=",
            ],
            "fs": [
                [
                    "list",
                    "fs_delete",
                    "fs_resize",
                    "fs_export",
                    "fs_snap_create",
                    "fs_snap_delete",
                    "fs_snap_restore",
                    "fs_dependants",
                    "fs_dependants_rm",
                    "file_clone",
                ],
                " --fs=",
            ],
            "nfs_export": [["list"], " --nfs-export="],
            "tgt": [["list"], " --tgt="],
            "job": [["job_status"], " --job="],
            "name": [
                [
                    "volume_create",
                    "volume_raid_create",
                    "volume_replicate",
                    "access_group_create",
                    "fs_create",
                    "fs_snap_create",
                ],
                " --name=",
            ],
            "size": [["volume_create", "volume_resize", "fs_create", "fs_resize"], " --size="],
            "provisioning": [["volume_create"], " --provisioning="],
            "raid_type": [["volume_raid_create"], " --raid-type="],
            "strip_size": [["volume_raid_create"], " --strip-size="],
            "rep_type": [["volume_replicate", "volume_replicate_range"], " --rep-type="],
            "src_vol": [["volume_replicate_range"], " --src-vol="],
            "dst_vol": [["volume_replicate_range"], " --dst-vol="],
            "src_start": [["volume_replicate_range"], " --src-start="],
            "dst_start": [["volume_replicate_range"], " --dst-start="],
            "count": [["volume_replicate_range"], " --count="],
            "read_pct": [["system_read_cache_pct_update"], " --read-pct="],
            "init": [["access_group_create", "access_group_add", "access_group_remove", "iscsi_chap"], " --init="],
            "in_user": [["iscsi_chap"], " --in-user="],
            "in_pass": [["iscsi_chap"], " --in-pass="],
            "out_user": [["iscsi_chap"], " --out-user="],
            "out_pass": [["iscsi_chap"], " --out-pass="],
            "export_path": [["iscsi_chap"], " --exportpath="],
            "anonuid": [["fs_export"], " --anonuid="],
            "anongid": [["fs_export"], " --anongid="],
            "auth_type": [["fs_export"], " --auth-type="],
            "root_host": [["fs_export"], " --root-host="],
            "ro_host": [["fs_export"], " --ro-host="],
            "rw_host": [["fs_export"], " --rw-host="],
            "export": [["fs_unexport"], " --export="],
            "src_fs": [["fs_clone"], " --src-fs="],
            "dst_name": [["fs_clone"], " --dst-name="],
            "backing_snapshot": [["fs_clone", "file_clone"], " --backing-snapshot="],
            "snap": [["fs_snap_delete", "fs_snap_restore"], " --snap="],
            "lsm_file": [["fs_snap_restore", "fs_dependants", "fs_dependants_rm"], " --file="],
            "fileas": [["fs_snap_restore"], " --fileas="],
            "src": [["file_clone"], " --src"],
            "dst": [["file_clone"], " --dst"],
            "policy": [
                ["volume_phy_disk_cache_update", "volume_read_cache_policy_update", "volume_read_cache_policy_update"],
                " --policy=",
            ],
            "path": [
                [
                    "local_disk_ident_led_on",
                    "local_disk_ident_led_off",
                    "local_disk_fault_led_on",
                    "local_disk_fault_led_on",
                ],
                " --path=",
            ],
        }

        Wrapper.__init__(self, self.commands, self.arguments, self.disable_check)

        _print("INFO: LSM configured")

    def _check(self, cmd):
        if self.disable_check or cmd:
            # Do not check if checking is disabled
            return True

        return True

    def _run(self, cmd, verbosity=True, return_output=False, **kwargs):
        # Constructs the command to run and runs it

        ret_fail = False
        if return_output:
            ret_fail = (False, None)

        try:
            command = self._add_command(cmd)
            command = self._add_arguments(command, **kwargs)

        except WrongCommandExceptionError as e:
            _print("WARN: Given command '%s' is not allowed in this version." % e.command)
            return ret_fail
        except WrongArgumentExceptionError as e:
            message = "WARN: Given argument '%s' is not allowed for given command." % e.argument
            if e.command:
                message = message[:-1] + " '" + e.command + "'."
            if e.arguments:
                message += "\nPlease use only these: %s." % ", ".join(e.arguments)
            _print(message)
            return ret_fail

        cmd = "lsmcli "
        if self.timeout:
            cmd += "-w %s " % self.timeout
        if self.protocol:
            cmd += '-u "%s://' % self.protocol
            if self.username and self.target:
                cmd += f"{self.username}@{self.target}"
            if self.port:
                cmd += ":%s" % self.port
            if self.query_params:
                cmd += "%s" % self.query_params
            cmd += '" '
        cmd += command

        try:
            self._check(cmd)
        except WrongArgumentExceptionError:
            pass
        except FailedCheckExceptionError as e:
            _print("WARN: Failed checking on argument %s" % e.argument)
            return ret_fail

        if return_output:
            ret, data = run(cmd, verbose=verbosity, return_output=True)
            if ret != 0:
                _print("WARN: Running command: '%s' failed. Return with output." % cmd)
            return ret, data

        ret = run(cmd, verbose=verbosity)
        if ret != 0:
            _print("WARN: Running command: '%s' failed." % cmd)
        return ret

    @staticmethod
    def _remove_nones(kwargs):
        return {k: v for k, v in kwargs.items() if v is not None}

    @_cli
    def list(
        self,
        lsm_type=None,
        fs=None,
        sys=None,
        pool=None,
        vol=None,
        disk=None,
        ag=None,
        nfs_export=None,
        tgt=None,
        **kwargs,
    ):
        kwargs.update(
            {
                "lsm_type": lsm_type,
                "fs": fs,
                "sys": sys,
                "pool": pool,
                "vol": vol,
                "disk": disk,
                "ag": ag,
                "nfs_export": nfs_export,
                "tgt": tgt,
            }
        )
        return self._run("list", **self._remove_nones(kwargs))

    @_cli
    def job_status(self, job=None, **kwargs):
        kwargs.update({"job": job})
        return self._run("job_status", **self._remove_nones(kwargs))

    @_cli
    def capabilities(self, sys=None, **kwargs):
        kwargs.update({"sys": sys})
        return self._run("capabilities", **self._remove_nones(kwargs))

    @_cli
    def plugin_info(self, **kwargs):
        return self._run("plugin_info", **self._remove_nones(kwargs))

    @_cli
    def volume_create(self, name=None, size=None, pool=None, provisioning=None, **kwargs):
        kwargs.update({"name": name, "size": size, "pool": pool, "provisioning": provisioning})
        return self._run("volume_create", **self._remove_nones(kwargs))

    @_cli
    def volume_raid_create(self, name=None, raid_type=None, disk=None, strip_size=None, **kwargs):
        kwargs.update({"name": name, "raid_type": raid_type, "disk": disk, "strip_size": strip_size})
        return self._run("volume_raid_create", **self._remove_nones(kwargs))

    @_cli
    def volume_raid_create_cap(self, sys=None, **kwargs):
        kwargs.update({"sys": sys})
        return self._run("volume_raid_create_cap", **self._remove_nones(kwargs))

    @_cli
    def volume_ident_led_on(self, vol=None, **kwargs):
        kwargs.update({"vol": vol})
        return self._run("volume_ident_led_on", **self._remove_nones(kwargs))

    @_cli
    def volume_ident_led_off(self, vol=None, **kwargs):
        kwargs.update({"vol": vol})
        return self._run("volume_ident_led_off", **self._remove_nones(kwargs))

    @_cli
    def volume_delete(self, vol=None, **kwargs):
        kwargs.update({"vol": vol})
        return self._run("volume_delete", **self._remove_nones(kwargs))

    @_cli
    def volume_resize(self, vol=None, size=None, **kwargs):
        kwargs.update({"vol": vol, "size": size})
        return self._run("volume_resize", **self._remove_nones(kwargs))

    @_cli
    def volume_replicate(self, vol=None, name=None, rep_type=None, pool=None, **kwargs):
        kwargs.update({"vol": vol, "name": name, "rep_type": rep_type, "pool": pool})
        return self._run("volume_replicate", **self._remove_nones(kwargs))

    @_cli
    def volume_replicate_range(
        self, src_vol=None, dst_vol=None, rep_type=None, src_start=None, dst_start=None, count=None, **kwargs
    ):
        kwargs.update(
            {
                "src_vol": src_vol,
                "dst_vol": dst_vol,
                "rep_type": rep_type,
                "src_start": src_start,
                "dst_start": dst_start,
                "count": count,
            }
        )
        return self._run("volume_replicate_range", **self._remove_nones(kwargs))

    @_cli
    def volume_replicate_range_block_size(self, sys=None, **kwargs):
        kwargs.update({"sys": sys})
        return self._run("volume_replicate_range_block_size", **self._remove_nones(kwargs))

    @_cli
    def volume_dependants(self, vol=None, **kwargs):
        kwargs.update({"vol": vol})
        return self._run("volume_dependants", **self._remove_nones(kwargs))

    @_cli
    def volume_dependants_rm(self, vol=None, **kwargs):
        kwargs.update({"vol": vol})
        return self._run("volume_dependants_rm", **self._remove_nones(kwargs))

    @_cli
    def volume_access_group(self, vol=None, **kwargs):
        kwargs.update({"vol": vol})
        return self._run("volume_access_group", **self._remove_nones(kwargs))

    @_cli
    def volume_mask(self, vol=None, ag=None, **kwargs):
        kwargs.update({"vol": vol, "ag": ag})
        return self._run("volume_mask", **self._remove_nones(kwargs))

    @_cli
    def volume_unmask(self, vol=None, ag=None, **kwargs):
        kwargs.update({"vol": vol, "ag": ag})
        return self._run("volume_unmask", **self._remove_nones(kwargs))

    @_cli
    def volume_enable(self, vol=None, **kwargs):
        kwargs.update({"vol": vol})
        return self._run("volume_enable", **self._remove_nones(kwargs))

    @_cli
    def volume_disable(self, vol=None, **kwargs):
        kwargs.update({"vol": vol})
        return self._run("volume_disable", **self._remove_nones(kwargs))

    @_cli
    def volume_raid_info(self, vol=None, **kwargs):
        kwargs.update({"vol": vol})
        return self._run("volume_raid_info", **self._remove_nones(kwargs))

    @_cli
    def pool_member_info(self, pool=None, **kwargs):
        kwargs.update({"pool": pool})
        return self._run("pool_member_info", **self._remove_nones(kwargs))

    @_cli
    def access_group_create(self, name=None, init=None, sys=None, **kwargs):
        kwargs.update({"name": name, "init": init, "sys": sys})
        return self._run("access_group_create", **self._remove_nones(kwargs))

    @_cli
    def access_group_add(self, ag=None, init=None, **kwargs):
        kwargs.update({"ag": ag, "init": init})
        return self._run("access_group_add", **self._remove_nones(kwargs))

    @_cli
    def access_group_remove(self, ag=None, init=None, **kwargs):
        kwargs.update({"ag": ag, "init": init})
        return self._run("access_group_remove", **self._remove_nones(kwargs))

    @_cli
    def access_group_delete(self, ag=None, **kwargs):
        kwargs.update({"ag": ag})
        return self._run("access_group_delete", **self._remove_nones(kwargs))

    @_cli
    def access_group_volumes(self, ag=None, **kwargs):
        kwargs.update({"ag": ag})
        return self._run("access_group_volumes", **self._remove_nones(kwargs))

    @_cli
    def iscsi_chap(self, init=None, in_user=None, in_pass=None, out_user=None, out_pass=None, **kwargs):
        kwargs.update(
            {"init": init, "in_user": in_user, "in_pass": in_pass, "out_user": out_user, "out_pass": out_pass}
        )
        return self._run("iscsi_chap", **self._remove_nones(kwargs))

    @_cli
    def fs_create(self, fs=None, **kwargs):
        kwargs.update({"fs": fs})
        return self._run("fs_create", **self._remove_nones(kwargs))

    @_cli
    def fs_delete(self, fs=None, **kwargs):
        kwargs.update({"fs": fs})
        return self._run("fs_delete", **self._remove_nones(kwargs))

    @_cli
    def fs_resize(self, fs=None, size=None, **kwargs):
        kwargs.update({"fs": fs, "size": size})
        return self._run("fs_resize", **self._remove_nones(kwargs))

    @_cli
    def fs_export(
        self,
        fs=None,
        exportpath=None,
        anonguid=None,
        anongid=None,
        auth_type=None,
        root_host=None,
        ro_host=None,
        rw_host=None,
        **kwargs,
    ):
        kwargs.update(
            {
                "fs": fs,
                "exportpath": exportpath,
                "anonguid": anonguid,
                "anongid": anongid,
                "auth_type": auth_type,
                "root_host": root_host,
                "ro_host": ro_host,
                "rw_host": rw_host,
            }
        )
        return self._run("fs_export", **self._remove_nones(kwargs))

    @_cli
    def fs_unexport(self, export=None, **kwargs):
        kwargs.update({"export": export})
        return self._run("fs_unexport", **self._remove_nones(kwargs))

    @_cli
    def fs_clone(self, src_fs=None, dst_name=None, backing_snapshot=None, **kwargs):
        kwargs.update({"src_fs": src_fs, "dst_name": dst_name, "backing_snapshot": backing_snapshot})
        return self._run("fs_clone", **self._remove_nones(kwargs))

    @_cli
    def fs_snap_create(self, name=None, fs=None, **kwargs):
        kwargs.update({"name": name, "fs": fs})
        return self._run("fs_snap_create", **self._remove_nones(kwargs))

    @_cli
    def fs_snap_delete(self, snap=None, fs=None, **kwargs):
        kwargs.update({"snap": snap, "fs": fs})
        return self._run("fs_snap_delete", **self._remove_nones(kwargs))

    @_cli
    def fs_snap_restore(self, fs=None, snap=None, lsm_file=None, fileas=None, **kwargs):
        kwargs.update({"fs": fs, "snap": snap, "lsm_file": lsm_file, "fileas": fileas})
        return self._run("fs_snap_restore", **self._remove_nones(kwargs))

    @_cli
    def fs_dependants(self, fs=None, lsm_file=None, **kwargs):
        kwargs.update({"fs": fs, "lsm_file": lsm_file})
        return self._run("fs_dependants", **self._remove_nones(kwargs))

    @_cli
    def fs_dependants_rm(self, fs=None, lsm_file=None, **kwargs):
        kwargs.update({"fs": fs, "lsm_file": lsm_file})
        return self._run("fs_dependants_rm", **self._remove_nones(kwargs))

    @_cli
    def file_clone(self, fs=None, src=None, dst=None, backing_snapshot=None, **kwargs):
        kwargs.update({"fs": fs, "src": src, "dst": dst, "backing_snapshot": backing_snapshot})
        return self._run("file_clone", **self._remove_nones(kwargs))

    @_cli
    def system_read_cache_pct_update(self, sys=None, read_pct=None, **kwargs):
        kwargs.update({"sys": sys, "read_pct": read_pct})
        return self._run("system_read_cache_pct_update", **self._remove_nones(kwargs))

    @_cli
    def local_disk_list(self, **kwargs):
        return self._run("local_disk_list", **self._remove_nones(kwargs))

    @_cli
    def volume_cache_info(self, vol=None, **kwargs):
        kwargs.update({"vol": vol})
        return self._run("volume_cache_info", **self._remove_nones(kwargs))

    @_cli
    def volume_phy_disk_cache_update(self, vol=None, policy=None, **kwargs):
        kwargs.update({"vol": vol, "policy": policy})
        return self._run("volume_phy_disk_cache_update", **self._remove_nones(kwargs))

    @_cli
    def volume_read_cache_policy_update(self, vol=None, policy=None, **kwargs):
        kwargs.update({"vol": vol, "policy": policy})
        return self._run("volume_read_cache_policy_update", **self._remove_nones(kwargs))

    @_cli
    def volume_write_cache_policy_update(self, vol=None, policy=None, **kwargs):
        kwargs.update({"vol": vol, "policy": policy})
        return self._run("volume_write_cache_policy_update", **self._remove_nones(kwargs))

    @_cli
    def local_disk_ident_led_on(self, path=None, **kwargs):
        kwargs.update({"path": path})
        return self._run("local_disk_ident_led_on", **self._remove_nones(kwargs))

    @_cli
    def local_disk_ident_led_off(self, path=None, **kwargs):
        kwargs.update({"path": path})
        return self._run("local_disk_ident_led_off", **self._remove_nones(kwargs))

    @_cli
    def local_disk_fault_led_on(self, path=None, **kwargs):
        kwargs.update({"path": path})
        return self._run("local_disk_fault_led_on", **self._remove_nones(kwargs))

    @_cli
    def local_disk_fault_led_off(self, path=None, **kwargs):
        kwargs.update({"path": path})
        return self._run("local_disk_fault_led_off", **self._remove_nones(kwargs))

    def help(self, cmd=""):
        """Retrieve help.
        The arguments are:
        \tcmd - optional | get help on this command
        Returns:
        \tBoolean:
        \t\tTrue if success
        \t\tFalse in case of failure
        """
        if cmd and cmd not in list(self.commands.keys()):
            _print("FAIL: Unknown command %s." % cmd)
            return False

        command = "%s -h" % cmd.replace("_", "-")
        return run(command, verbose=True)

    def version(self, cmd=""):
        """Retrieve plugin version.
        The arguments are:
        \tcmd - optional | get version of this command
        Returns:
        \tBoolean:
        \t\tTrue if success
        \t\tFalse in case of failure
        """
        if cmd and cmd not in list(self.commands.keys()):
            _print("FAIL: Unknown command %s." % cmd)
            return False

        command = "%s -v" % cmd.replace("_", "-")
        return run(command, verbose=True)



#!/usr/bin/python

# Copyright (C) 2016 Red Hat, Inc.
# This file is part of libsan.
#
# libsan is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# libsan is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with libsan.  If not, see <http://www.gnu.org/licenses/>.

import os

import pytest
from libsan.host import fio
from libsan.host.cmdline import run


def test_fio():
    if not fio.install_fio():
        pytest.fail("Unable to install fio")

    output_file = "fio.test"

    # Create a file big enough to use by FIO
    run("dd if=/dev/zero of=%s count=20 bs=1024k" % output_file)

    try:
        fio_pid = fio.fio_stress_background(output_file, size="1m")
    except Exception as e:
        pytest.fail("FAIL: Exception: %s" % e)

    # Make sure background process does not generate exception
    print("INFO: Waiting FIO process to finish")
    try:
        _, exit_status = os.waitpid(fio_pid, 0)
    except Exception as e:
        pytest.fail("FAIL: Exception: %s" % e)

    if exit_status != 0:
        pytest.fail("FAIL: there was some error running FIO")
    try:
        os.remove(output_file)
    except Exception:
        pytest.fail("FAIL: Could not delete %s" % output_file)

    assert 1



# Copyright (C) 2016 Red Hat, Inc.
# This file is part of libsan.
#
# libsan is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# libsan is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with libsan.  If not, see <http://www.gnu.org/licenses/>.

"""array.py: Module to manage Arrays or Lists."""

__author__ = "Bruno Goncalves"
__copyright__ = "Copyright (c) 2016 Red Hat, Inc. All rights reserved."


def dedup(array):
    """Remove duplicated entries from a list"""
    if not array:
        return None
    # order preserving
    checked = []
    for e in array:
        if e not in checked:
            checked.append(e)
    return checked


def is_array_same(list1, list2):
    """Check if 2 unordered lists have the same elements"""
    if set(list1) == set(list2):
        return True
    return False


def lowest_free_number(list_numbers):
    """From a list of integers return the lowest available number
    Start with 0,
    Example:
        input: [0, 1, 2, 4, 5, 6]
        ouput: 3
        input: [0, 1, 2, 3, 4, 5, 6]
        ouput: 7
    """
    if not list_numbers:
        return 0

    int_list = []
    # if items are not int, try to convert them to int
    for entry in list_numbers:
        int_list.append(int(entry))

    size_list = len(int_list)
    for i in range(size_list):
        if i not in int_list:
            return i

    return size_list



# Copyright (C) 2018 Red Hat, Inc.
# This file is part of libsan.
#
# libsan is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# libsan is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with libsan.  If not, see <http://www.gnu.org/licenses/>.

"""dmpd.py: Module to manipulate LVM thinp and cache metadata devices and snapshot eras."""

__author__ = "Jakub Krysl"
__copyright__ = "Copyright (c) 2017 Red Hat, Inc. All rights reserved."

import os
import sys

import libsan.host.linux
import libsan.host.lvm
from libsan import _print
from libsan.host.cli_tools import Wrapper
from libsan.host.cmdline import run


def _get_devices():
    lv_list = libsan.host.lvm.lv_query()
    return lv_list


def _get_active_devices():
    cmd = "ls /dev/mapper/"
    retcode, output = run(cmd, return_output=True, verbose=False)
    if retcode != 0:
        _print("FAIL: Could not find active dm devices")
        return False
    devices = output.split()
    return devices


def _get_device_path(vg_name, lv_name):
    device_path = vg_name + "-" + lv_name
    if "/dev/mapper/" not in device_path:
        device_path = "/dev/mapper/" + device_path
    return device_path


def _check_device(vg_name, lv_name):
    devices = _get_devices()
    device_list = [f["name"] for f in devices]
    if lv_name not in device_list:
        _print("FAIL: %s is not a device" % lv_name)
        return False
    for x in devices:
        if x["name"] == lv_name and x["vg_name"] == vg_name:
            _print(f"INFO: Found device {lv_name} in group {vg_name}")
            return True
    return False


def _activate_device(vg_name, lv_name):
    devices_active = _get_active_devices()
    if vg_name + "-" + lv_name not in devices_active:
        ret = libsan.host.lvm.lv_activate(lv_name, vg_name)
        if not ret:
            _print("FAIL: Could not activate device %s" % lv_name)
            return False
        _print("INFO: device %s was activated" % lv_name)
    _print("INFO: device %s is active" % lv_name)
    return True


def _fallocate(_file, size, command_message):
    cmd = f"fallocate -l {size}M {_file}"
    try:
        retcode = run(cmd)
        if retcode != 0:
            _print("FAIL: Command failed with code %s." % retcode)
            _print("FAIL: Could not create file to %s metadata to." % command_message)
            return False
    except OSError as e:
        print("command failed: ", e, file=sys.stderr)
        return False
    return True


def get_help(cmd):
    commands = [
        "cache_check",
        "cache_dump",
        "cache_metadata_size",
        "cache_repair",
        "cache_restore",
        "era_check",
        "era_dump",
        "era_invalidate",
        "era_restore",
        "thin_check",
        "thin_delta",
        "thin_dump",
        "thin_ls",
        "thin_metadata_size",
        "thin_repair",
        "thin_restore",
        "thin_rmap",
        "thin_show_duplicates",
        "thin_trim",
    ]
    if cmd not in commands:
        _print("FAIL: Unknown command %s" % cmd)
        return False

    command = "%s -h" % cmd
    retcode = run(command, verbose=True)
    if retcode != 0:
        _print("FAIL: Could not get help for %s." % cmd)
        return False

    return True


def get_version(cmd):
    commands = [
        "cache_check",
        "cache_dump",
        "cache_metadata_size",
        "cache_repair",
        "cache_restore",
        "era_check",
        "era_dump",
        "era_invalidate",
        "era_restore",
        "thin_check",
        "thin_delta",
        "thin_dump",
        "thin_ls",
        "thin_metadata_size",
        "thin_repair",
        "thin_restore",
        "thin_rmap",
        "thin_show_duplicates",
        "thin_trim",
    ]
    if cmd not in commands:
        _print("FAIL: Unknown command %s" % cmd)
        return False

    command = "%s -V" % cmd
    retcode = run(command, verbose=True)
    if retcode != 0:
        _print("FAIL: Could not get version of %s." % cmd)
        return False

    return True


def _get_dev_id(dev_id, path=None, lv_name=None, vg_name=None):
    dev_ids = []

    if path is None:
        retcode, data = thin_dump(source_vg=vg_name, source_lv=lv_name, formatting="xml", return_output=True)
        if not retcode:
            _print(f"FAIL: Could not dump metadata from {vg_name}/{lv_name}")
            return False
        data_lines = data.splitlines()
        for line in data_lines:
            blocks = line.split()
            for block in blocks:
                if not block.startswith("dev_"):
                    continue
                dev_ids.append(int(block[8:-1]))

    else:
        with open(path, encoding="UTF-8") as meta:
            for line in meta:
                blocks = line.split()
                for block in blocks:
                    if not block.startswith("dev_"):
                        continue
                    dev_ids.append(int(block[8:-1]))

    if dev_id in dev_ids:
        return True

    return False


def _metadata_size(source=None, lv_name=None, vg_name=None):
    if source is None:
        cmd = "lvs -a --units m"
        ret, data = run(cmd, return_output=True)
        if ret != 0:
            _print("FAIL: Could not list LVs")
        data_line = data.splitlines()
        for line in data_line:
            cut = line.split()
            if not cut or lv_name != cut[0] and vg_name != cut[1]:
                continue
            cut = cut[3]
            cut = cut.split("m")
            size = float(cut[0])
            run(cmd)
            return int(size)
        _print(f"FAIL: Could not find {lv_name} {vg_name} in lvs, setting size to 100m")
        return 100
    else:
        return int(os.stat(source).st_size) / 1000000


###########################################
# cache section
###########################################


def cache_check(
    source_file=None,
    source_vg=None,
    source_lv=None,
    quiet=False,
    super_block_only=False,
    clear_needs_check_flag=False,
    skip_mappings=False,
    skip_hints=False,
    skip_discards=False,
    verbose=True,
):
    """Check cache pool metadata from either file or device.
    The arguments are:
    \tsource_file
    \tsource_vg VG name
    \tsource_lv LV name
    \tquiet Mute STDOUT
    \tsuper_block_only
    \tclear_needs_check_flag
    \tskip_mappings
    \tskip_hints
    \tskip_discards
    Returns:
    \tBoolean:
    \t\tTrue if success
    \t'tFalse in case of failure
    """

    options = ""

    if not source_file and (not source_vg or not source_lv):
        _print("FAIL: cache_check requires either source_file OR source_vg and source_lv.")
        return False

    if not source_file:
        ret = _check_device(source_vg, source_lv)
        if not ret:
            return False
        ret = _activate_device(source_vg, source_lv)
        if not ret:
            return False
        device = _get_device_path(source_vg, source_lv)
    else:
        if not os.path.isfile(source_file):
            _print("FAIL: Source file is not a file.")
            return False
        device = source_file

    if quiet:
        options += "--quiet "

    if super_block_only:
        options += "--super-block-only "

    if clear_needs_check_flag:
        options += "--clear-needs-check-flag "

    if skip_mappings:
        options += "--skip-mappings "

    if skip_hints:
        options += "--skip-hints "

    if skip_discards:
        options += "--skip-discards "

    cmd = f"cache_check {device} {options}"
    retcode = run(cmd, verbose=verbose)
    if retcode != 0:
        _print("FAIL: Could not check %s metadata" % device)
        return False

    return True


def cache_dump(
    source_file=None, source_vg=None, source_lv=None, output=None, repair=False, verbose=True, return_output=False
):
    """Dumps cache metadata from device of source file to standard output or file.
    The arguments are:
    \tsource_file
    \tsource_vg VG name
    \tsource_lv LV name
    \toutput specify output xml file
    \treturn_output see 'Returns', not usable with output=True
    \trepair Repair the metadata while dumping it
    Returns:
    \tOnly Boolean if return_output False:
    \t\tTrue if success
    \t'tFalse in case of failure
    \tBoolean and data if return_output True
    """
    options = ""
    data = None

    if return_output and output:
        _print("INFO: Cannot return to both STDOUT and file, returning only to file.")
        return_output = False

    ret_fail = (False, None) if return_output else False

    if not source_file and (not source_vg or not source_lv):
        _print("FAIL: cache_dump requires either source_file OR source_vg and source_lv.")
        return ret_fail

    if not source_file:
        ret = _check_device(source_vg, source_lv)
        if not ret:
            return ret_fail
        ret = _activate_device(source_vg, source_lv)
        if not ret:
            return ret_fail
        device = _get_device_path(source_vg, source_lv)
    else:
        if not os.path.isfile(source_file):
            _print("FAIL: Source file is not a file.")
            return ret_fail
        device = source_file

    if output:
        if not os.path.isfile(output):
            size = _metadata_size(source_file, source_lv, source_vg)
            ret = _fallocate(output, size + 1, "dump")
            if not ret:
                return ret_fail
        options += "-o %s " % output

    if repair:
        options += "--repair"

    cmd = f"cache_dump {device} {options}"
    if return_output:
        retcode, data = run(cmd, return_output=True, verbose=verbose)
    else:
        retcode = run(cmd, verbose=verbose)
    if retcode != 0:
        _print("FAIL: Could not dump %s metadata." % device)
        return ret_fail

    if return_output:
        return True, data
    return True


def cache_repair(
    source_file=None, source_vg=None, source_lv=None, target_file=None, target_vg=None, target_lv=None, verbose=True
):
    """Repairs cache metadata from source file/device to target file/device
    The arguments are:
    \tsource as either source_file OR source_vg and source_lv
    \ttarget as either target_file OR target_vg and target_lv
    Returns:
    \tBoolean:
    \t\tTrue if success
    \t'tFalse in case of failure
    """

    if not source_file and (not source_vg or not source_lv):
        _print("FAIL: cache_repair requires either source_file OR source_vg and source_lv as source.")
        return False

    if not target_file and (not target_vg or not target_lv):
        _print("FAIL: cache_repair requires either target_file OR target_vg and target_lv as target.")
        return False

    if not source_file:
        ret = _check_device(source_vg, source_lv)
        if not ret:
            return False
        ret = _activate_device(source_vg, source_lv)
        if not ret:
            return False
        source = _get_device_path(source_vg, source_lv)
    else:
        if not os.path.isfile(source_file):
            _print("FAIL: Source file is not a file.")
            return False
        source = source_file

    if not target_file:
        ret = _check_device(target_vg, target_lv)
        if not ret:
            return False
        ret = _activate_device(target_vg, target_lv)
        if not ret:
            return False
        target = _get_device_path(target_vg, target_lv)
    else:
        if not os.path.isfile(target_file):
            size = _metadata_size(source_file, source_lv, source_vg)
            ret = _fallocate(target_file, size + 1, "repair")
            if not ret:
                return False
        target = target_file

    cmd = f"cache_repair -i {source} -o {target}"
    retcode = run(cmd, verbose=verbose)
    if retcode != 0:
        _print(f"FAIL: Could not repair metadata from {source} to {target}")
        return False

    return True


def cache_restore(
    source_file,
    target_vg=None,
    target_lv=None,
    target_file=None,
    quiet=False,
    metadata_version=None,
    omit_clean_shutdown=False,
    override_metadata_version=None,
    verbose=True,
):
    """Restores cache metadata from source xml file to target device/file
    The arguments are:
    \tsource_file Source xml file
    \ttarget as either target_file OR target_vg and target_lv
    \tquiet Mute STDOUT
    \tmetadata_version Specify metadata version to restore
    \tomit_clean_shutdown Disable clean shutdown
    \toverride_metadata_version DEBUG option to override metadata version without checking
    Returns:
    \tBoolean:
    \t\tTrue if success
    \t'tFalse in case of failure
    """

    options = ""

    if source_file is None:
        _print("FAIL: cache_restore requires source file.")
        return False

    if not target_file and (not target_vg or not target_lv):
        _print("FAIL: cache_restore requires either target_file OR target_vg and target_lv as target.")
        return False

    if not os.path.isfile(source_file):
        _print("FAIL: Source file is not a file.")
        return False

    if not target_file:
        ret = _check_device(target_vg, target_lv)
        if not ret:
            return False
        ret = _activate_device(target_vg, target_lv)
        if not ret:
            return False
        target = _get_device_path(target_vg, target_lv)
    else:
        if not os.path.isfile(target_file):
            size = _metadata_size(source_file)
            ret = _fallocate(target_file, size + 1, "restore")
            if not ret:
                return False
        target = target_file

    if quiet:
        options += "--quiet "

    if metadata_version:
        options += "--metadata-version %s " % metadata_version

    if omit_clean_shutdown:
        options += "--omit-clean-shutdown "

    if override_metadata_version:
        options += "--debug-override-metadata-version %s" % override_metadata_version

    cmd = f"cache_restore -i {source_file} -o {target} {options}"

    retcode = run(cmd, verbose=verbose)
    if retcode != 0:
        _print(f"FAIL: Could not restore metadata from {source_file} to {target}")
        return False

    return True


###########################################
# thinp section
###########################################


def thin_check(
    source_file=None,
    source_vg=None,
    source_lv=None,
    quiet=False,
    super_block_only=False,
    clear_needs_check_flag=False,
    skip_mappings=False,
    ignore_non_fatal_errors=False,
    verbose=True,
):
    """Check thin pool metadata from either file or device.
    The arguments are:
    \tsource_file
    \tsource_vg VG name
    \tsource_lv LV name
    \tquiet Mute STDOUT
    \tsuper_block_only
    \tclear_needs_check_flag
    \tskip_mappings
    \tignore_non_fatal_errors
    Returns:
    \tBoolean:
    \t\tTrue if success
    \t'tFalse in case of failure
    """

    options = ""

    if not source_file and (not source_vg or not source_lv):
        _print("FAIL: thin_check requires either source_file OR source_vg and source_lv.")
        return False

    if not source_file:
        ret = _check_device(source_vg, source_lv)
        if not ret:
            return False
        ret = _activate_device(source_vg, source_lv)
        if not ret:
            return False
        device = _get_device_path(source_vg, source_lv)
    else:
        if not os.path.isfile(source_file):
            _print("FAIL: Source file is not a file.")
            return False
        device = source_file

    if quiet:
        options += "--quiet "

    if super_block_only:
        options += "--super-block-only "

    if clear_needs_check_flag:
        options += "--clear-needs-check-flag "

    if skip_mappings:
        options += "--skip-mappings "

    if ignore_non_fatal_errors:
        options += "--ignore-non-fatal-errors "

    cmd = f"thin_check {device} {options}"
    retcode = run(cmd, verbose=verbose)
    if retcode != 0:
        _print("FAIL: Could not check %s metadata" % device)
        return False

    return True


def thin_ls(source_vg, source_lv, no_headers=False, fields=None, snapshot=False, verbose=True):
    """List information about thin LVs on thin pool.
    The arguments are:
    \tsource_vg VG name
    \tsource_lv LV name
    \tfields list of fields to output, default is all
    \tsnapshot If use metadata snapshot, able to run on live snapshotted pool
    Returns:
    \tBoolean:
    \t\tTrue if success
    \t'tFalse in case of failure
    """

    options = ""

    if not source_vg or not source_lv:
        _print("FAIL: thin_ls requires source_vg and source_lv.")
        return False

    ret = _check_device(source_vg, source_lv)
    if not ret:
        return False
    ret = _activate_device(source_vg, source_lv)
    if not ret:
        return False
    device = _get_device_path(source_vg, source_lv)

    if no_headers:
        options += "--no-headers "

    fields_possible = [
        "DEV",
        "MAPPED_BLOCKS",
        "EXCLUSIVE_BLOCKS",
        "SHARED_BLOCKS",
        "MAPPED_SECTORS",
        "EXCLUSIVE_SECTORS",
        "SHARED_SECTORS",
        "MAPPED_BYTES",
        "EXCLUSIVE_BYTES",
        "SHARED_BYTES",
        "MAPPED",
        "EXCLUSIVE",
        "TRANSACTION",
        "CREATE_TIME",
        "SHARED",
        "SNAP_TIME",
    ]
    if fields is None:
        options += ' --format "%s" ' % ",".join([str(i) for i in fields_possible])
    else:
        for field in fields:
            if field not in fields_possible:
                _print("FAIL: Unknown field %s specified." % field)
                _print("INFO: Possible fields are: %s" % ", ".join([str(i) for i in fields_possible]))
                return False
        options += ' --format "%s" ' % ",".join([str(i) for i in fields])

    if snapshot:
        options += "--metadata-snap"

    cmd = f"thin_ls {device} {options}"
    retcode = run(cmd, verbose=verbose)
    if retcode != 0:
        _print("FAIL: Could not list %s metadata" % device)
        return False

    return True


def thin_dump(
    source_file=None,
    source_vg=None,
    source_lv=None,
    output=None,
    repair=False,
    formatting=None,
    snapshot=None,
    dev_id=None,
    skip_mappings=False,
    verbose=True,
    return_output=False,
):
    """Dumps thin metadata from device of source file to standard output or file.
    The arguments are:
    \tsource_file
    \tsource_vg VG name
    \tsource_lv LV name
    \toutput specify output xml file
    \treturn_output see 'Returns', not usable with output=True
    \trepair Repair the metadata while dumping it
    \tformatting Specify output format [xml, human_readable, custom='file']
    \tsnapshot (Boolean/Int) Use metadata snapshot. If Int provided, specifies block number
    \tdev_id ID of the device
    Returns:
    \tOnly Boolean if return_output False:
    \t\tTrue if success
    \t'tFalse in case of failure
    \tBoolean and data if return_output True
    """
    options = ""
    data = None

    if return_output and output:
        _print("INFO: Cannot return to both STDOUT and file, returning only to file.")
        return_output = False

    ret_fail = (False, None) if return_output else False

    if not source_file and (not source_vg or not source_lv):
        _print("FAIL: thin_dump requires either source_file OR source_vg and source_lv.")
        return ret_fail

    if not source_file:
        ret = _check_device(source_vg, source_lv)
        if not ret:
            return ret_fail
        ret = _activate_device(source_vg, source_lv)
        if not ret:
            return ret_fail
        device = _get_device_path(source_vg, source_lv)
    else:
        if not os.path.isfile(source_file):
            _print("FAIL: Source file is not a file.")
            return ret_fail
        device = source_file

    if output:
        if not os.path.isfile(output):
            size = _metadata_size(source_file, source_lv, source_vg)
            ret = _fallocate(output, size + 1, "dump")
            if not ret:
                return ret_fail
        options += "-o %s " % output

    if repair:
        options += "--repair "

    if snapshot:
        if isinstance(snapshot, bool):
            options += "--metadata-snap "
        elif isinstance(snapshot, int):
            options += "--metadata-snap %s " % snapshot
        else:
            _print("FAIL: Unknown snapshot value, use either Boolean or Int.")
            return ret_fail

    if formatting:
        if formatting in ["xml", "human_readable"]:
            options += "--format %s " % formatting
        elif formatting.startswith("custom="):
            if not os.path.isfile(formatting[8:-1]):
                _print("FAIL: Specified custom formatting file is not a file.")
                return ret_fail
            options += "--format %s " % formatting
        else:
            _print("FAIL: Unknown formatting specified, please use one of [xml, human_readable, custom='file'].")
            return ret_fail

    if dev_id:
        if isinstance(dev_id, int):
            if _get_dev_id(dev_id, source_file, source_lv, source_vg):
                options += "--dev-id %s " % dev_id
            else:
                _print("FAIL: Unknown dev_id value, device with ID %s does not exist." % dev_id)
                return ret_fail
        else:
            _print("FAIL: Unknown dev_id value, must be Int.")
            return ret_fail

    if skip_mappings:
        options += "--skip-mappings "

    cmd = f"thin_dump {device} {options}"
    if return_output:
        retcode, data = run(cmd, return_output=True, verbose=verbose)
    else:
        retcode = run(cmd, verbose=verbose)
    if retcode != 0:
        _print("FAIL: Could not dump %s metadata." % device)
        return ret_fail

    if return_output:
        return True, data
    return True


def thin_restore(source_file, target_vg=None, target_lv=None, target_file=None, quiet=False, verbose=True):
    """Restores thin metadata from source xml file to target device/file
    The arguments are:
    \tsource_file Source xml file
    \ttarget as either target_file OR target_vg and target_lv
    \tquiet Mute STDOUT
    \tmetadata_version Specify metadata version to restore
    \tomit_clean_shutdown Disable clean shutdown
    \toverride_metadata_version DEBUG option to override metadata version without checking
    Returns:
    \tBoolean:
    \t\tTrue if success
    \t'tFalse in case of failure
    """

    options = ""

    if source_file is None:
        _print("FAIL: thin_restore requires source file.")
        return False

    if not target_file and (not target_vg or not target_lv):
        _print("FAIL: thin_restore requires either target_file OR target_vg and target_lv as target.")
        return False

    if not os.path.isfile(source_file):
        _print("FAIL: Source file is not a file.")
        return False

    if not target_file:
        ret = _check_device(target_vg, target_lv)
        if not ret:
            return False
        ret = _activate_device(target_vg, target_lv)
        if not ret:
            return False
        target = _get_device_path(target_vg, target_lv)
    else:
        if not os.path.isfile(target_file):
            size = _metadata_size(source_file)
            ret = _fallocate(target_file, size + 1, "restore")
            if not ret:
                return False
        target = target_file

    if quiet:
        options += "--quiet"

    cmd = f"thin_restore -i {source_file} -o {target} {options}"

    retcode = run(cmd, verbose=verbose)
    if retcode != 0:
        _print(f"FAIL: Could not restore metadata from {source_file} to {target}")
        return False

    return True


def thin_repair(
    source_file=None, source_vg=None, source_lv=None, target_file=None, target_vg=None, target_lv=None, verbose=True
):
    """Repairs thin metadata from source file/device to target file/device
    The arguments are:
    \tsource as either source_file OR source_vg and source_lv
    \ttarget as either target_file OR target_vg and target_lv
    Returns:
    \tBoolean:
    \t\tTrue if success
    \t'tFalse in case of failure
    """

    if not source_file and (not source_vg or not source_lv):
        _print("FAIL: thin_repair requires either source_file OR source_vg and source_lv as source.")
        return False

    if not target_file and (not target_vg or not target_lv):
        _print("FAIL: thin_repair requires either target_file OR target_vg and target_lv as target.")
        return False

    if not source_file:
        ret = _check_device(source_vg, source_lv)
        if not ret:
            return False
        ret = _activate_device(source_vg, source_lv)
        if not ret:
            return False
        source = _get_device_path(source_vg, source_lv)
    else:
        if not os.path.isfile(source_file):
            _print("FAIL: Source file is not a file.")
            return False
        source = source_file

    if not target_file:
        ret = _check_device(target_vg, target_lv)
        if not ret:
            return False
        ret = _activate_device(target_vg, target_lv)
        if not ret:
            return False
        target = _get_device_path(target_vg, target_lv)
    else:
        if not os.path.isfile(target_file):
            size = _metadata_size(source_file, source_lv, source_vg)
            ret = _fallocate(target_file, size + 1, "repair")
            if not ret:
                return False
        target = target_file

    cmd = f"thin_repair -i {source} -o {target}"
    retcode = run(cmd, verbose=verbose)
    if retcode != 0:
        _print(f"FAIL: Could not repair metadata from {source} to {target}")
        return False

    return True


def thin_rmap(region, source_file=None, source_vg=None, source_lv=None, verbose=True):
    """Output reverse map of a thin provisioned region of blocks from metadata device.
    The arguments are:
    \tsource_vg VG name
    \tsource_lv LV name
    Returns:
    \tBoolean:
    \t\tTrue if success
    \t'tFalse in case of failure
    """

    if not source_file and (not source_vg or not source_lv):
        _print("FAIL: thin_rmap requires either source_file OR source_vg and source_lv as source.")
        return False

    if not source_file:
        ret = _check_device(source_vg, source_lv)
        if not ret:
            return False
        ret = _activate_device(source_vg, source_lv)
        if not ret:
            return False
        device = _get_device_path(source_vg, source_lv)
    else:
        if not os.path.isfile(source_file):
            _print("FAIL: Source file is not a file.")
            return False
        device = source_file

    regions = region.split(".")
    try:
        int(regions[0])
        if regions[1] != "":
            raise ValueError  # noqa: TRY301
        int(regions[2])
        if regions[3] is not None:
            raise ValueError  # noqa: TRY301
    except ValueError:
        print("FAIL: Region must be in format 'INT..INT'")
        return False
    except IndexError:
        pass
    # region 1..-1 must be valid, using usigned 32bit ints
    if int(regions[0]) & 0xFFFFFFFF >= int(regions[2]) & 0xFFFFFFFF:
        print("FAIL: Beginning of the region must be before its end.")
        return False
    options = "--region %s" % region

    cmd = f"thin_rmap {device} {options}"
    retcode = run(cmd, verbose=verbose)
    if retcode != 0:
        _print("FAIL: Could not output reverse map from %s metadata device" % device)
        return False

    return True


def thin_trim(data_vg, data_lv, metadata_vg=None, metadata_lv=None, metadata_file=None, verbose=True):
    """Issue discard requests for free pool space.
    The arguments are:
    \tdata_vg VG name of data device
    \tdata_lv LV name of data device
    \tmetadata_vg VG name of metadata device
    \t metadata_lv LV name of metadata device
    \tmetadata_file file with metadata
    Returns:
    \tBoolean:
    \t\tTrue if success
    \t'tFalse in case of failure
    """
    options = ""

    if not data_vg or not data_lv:
        _print("FAIL: thin_trim requires data_vg and data_lv.")
        return False

    if not metadata_file and (not metadata_vg or not metadata_lv):
        _print("FAIL: thin_trim requires either metadata_file OR metadata_vg and metadata_lv as target.")
        return False

    ret = _check_device(data_vg, data_lv)
    if not ret:
        return False

    ret = _activate_device(data_vg, data_lv)
    if not ret:
        return False

    if not metadata_file:
        ret = _check_device(metadata_vg, metadata_lv)
        if not ret:
            return False
        ret = _activate_device(metadata_vg, metadata_lv)
        if not ret:
            return False
        metadata_dev = _get_device_path(metadata_vg, metadata_lv)
    else:
        if not os.path.isfile(metadata_file):
            _print("FAIL: metadata_file %s is not a file." % metadata_file)
            return False
        metadata_dev = metadata_file

    data_dev = _get_device_path(data_vg, data_lv)
    cmd = f"thin_trim --data-dev {data_dev} --metadata-dev {metadata_dev} {options}"
    retcode = run(cmd, verbose=verbose)
    if retcode != 0:
        _print(f"FAIL: Could not discard free pool space on device {data_dev} with metadata device {metadata_dev}.")
        return False

    return True


def thin_delta(
    thin1, thin2, source_file=None, source_vg=None, source_lv=None, snapshot=False, verbosity=False, verbose=True
):
    """Print the differences in the mappings between two thin devices..
    The arguments are:
    \tsource_vg VG name
    \tsource_lv LV name
    \tthin1 numeric identificator of first thin volume
    \tthin2 numeric identificator of second thin volume
    \tsnapshot (Boolean/Int) Use metadata snapshot. If Int provided, specifies block number
    \tverbosity Provide extra information on the mappings
    Returns:
    \tBoolean:
    \t\tTrue if success
    \t'tFalse in case of failure
    """

    options = ""

    if not source_file and (not source_vg or not source_lv):
        _print("FAIL: thin_delta requires either source_file OR source_vg and source_lv.")
        return False

    if not source_file:
        ret = _check_device(source_vg, source_lv)
        if not ret:
            return False
        ret = _activate_device(source_vg, source_lv)
        if not ret:
            return False
        device = _get_device_path(source_vg, source_lv)
    else:
        if not os.path.isfile(source_file):
            _print("FAIL: Source file is not a file.")
            return False
        device = source_file

    if snapshot:
        if isinstance(snapshot, bool):
            options += "--metadata-snap "
        elif isinstance(snapshot, int):
            options += "--metadata-snap %s " % snapshot
        else:
            _print("FAIL: Unknown snapshot value, use either Boolean or Int.")
            return False

    if verbosity:
        options += "--verbose"

    if _get_dev_id(thin1, source_file, source_lv, source_vg) and _get_dev_id(thin2, source_file, source_lv, source_vg):
        cmd = f"thin_delta {options} --thin1 {thin1} --thin2 {thin2} {device}"
        retcode = run(cmd, verbose=verbose)
        if retcode != 0:
            _print("FAIL: Could not get differences in mappings between two thin LVs.")
            return False
    else:
        _print("FAIL: Specified ID does not exist.")
        return False
    return True


class DMPD(Wrapper):
    def __init__(self, disable_check=True):
        self.disable_check = disable_check

        pkg = "device-mapper-persistent-data"
        if not libsan.host.linux.is_installed(pkg) and not libsan.host.linux.install_package(pkg, check=False):
            _print("FATAL: Could not install %s package" % pkg)

        self.commands = {
            "cache_check": "cache_check",
            "cache_dump": "cache_dump",
            "cache_metadata_size": "cache_metadata_size",
            "cache_repair": "cache_repair",
            "cache_restore": "cache_restore",
            "cache_writeback": "cache_writeback",
            "thin_check": "thin_check",
            "thin_delta": "thin_delta",
            "thin_dump": "thin_dump",
            "thin_ls": "thin_ls",
            "thin_metadata_size": "thin_metadata_size",
            "thin_repair": "thin_repair",
            "thin_restore": "thin_restore",
            "thin_rmap": "thin_rmap",
            "thin_trim": "thin_trim",
        }
        self.commands["all"] = list(self.commands.keys())
        self.arguments = {
            "help": [self.commands["all"], " --help"],
            "version": [self.commands["all"], " --version"],
            "verbose": [self.commands["all"], " --verbose"],
            "block_size": [["cache_metadata_size", "thin_metadata_size"], " --block-size&"],
            "buffer_size": [["cache_writeback"], " --buffer-size-meg&"],
            "clear_needs_check_flag": [["cache_check", "thin_check"], " --clear-needs-check-flag"],
            "data_dev": [["thin_trim"], " --data-dev&"],
            "debug_override_metadata_version": [["cache_restore"], " --debug-override-metadata-version&"],
            "dev_id": [["thin_dump"], " --dev-id&"],
            "device_size": [["cache_metadata_size"], " --device-size&"],
            "fast_device": [["cache_writeback"], " --fast-device&"],
            "format": [["thin_ls", "thin_dump"], " --format&"],
            "ignore_non_fatal_errors": [["thin_check"], " --ignore-non-fatal-errors"],
            "input": [["cache_repair", "cache_restore", "thin_repair", "thin_restore"], " -i&"],
            "list_failed_blocks": [["cache_writeback"], " --list-failed-blocks"],
            "max_hint_width": [["cache_metadata_size"], " --max-hint-width&"],
            "max_thins": [["thin_metadata_size"], " --max-thins&"],
            "metadata_dev": [["thin_trim"], " --metadata-dev&"],
            "metadata_device": [["cache_writeback"], " --metadata-device&"],
            "metadata_snap": [["thin_dump", "thin_delta"], " --metadata-snap&"],
            "metadata_version": [["cache_restore"], " --metadata-version&"],
            "no_headers": [["thin_ls"], " --no-headers"],
            "no_metadata_update": [["cache_writeback"], " --no-metadata-update"],
            "nr_blocks": [["cache_metadata_size"], " --nr-blocks&"],
            "numeric_only": [["thin_metadata_size"], " --numeric-only"],
            "numeric_only_type": [["thin_metadata_size"], " --numeric-only="],
            "omit_clean_shutdown": [["cache_restore"], " --omit-clean-shutdown"],
            "origin_device": [["cache_writeback"], " --origin-device&"],
            "output": [
                ["cache_dump", "cache_repair", "cache_restore", "thin_dump", "thin_repair", "thin_restore"],
                " -o&",
            ],
            "override_mapping_root": [["thin_check"], " --override-mapping-root&"],
            "pool_size": [["thin_metadata_size"], " --pool-size&"],
            "quiet": [["cache_check", "cache_restore", "thin_check", "thin_restore"], " --quiet"],
            "region": [["thin_rmap"], " --region&"],
            "repair": [["cache_dump", "thin_dump"], " --repair"],
            "skip_discards": [["cache_check"], " --skip-discards"],
            "skip_hints": [["cache_check"], " --skip-hints"],
            "skip_mappings": [["thin_check", "thin_dump"], " --skip-mappings"],
            "snap1": [["thin_delta"], " --snap1&"],
            "snap2": [["thin_delta"], " --snap2&"],
            "snapshot": [["thin_ls", "thin_dump", "thin_delta"], " --metadata-snap"],
            "super_block_only": [["cache_check", "thin_check"], " --super-block-only"],
            "thin1": [["thin_delta"], " --thin1&"],
            "thin2": [["thin_delta"], " --thin2&"],
            "unit": [["thin_metadata_size"], " --unit&"],
        }

        Wrapper.__init__(self, self.commands, self.arguments, self.disable_check)

    @staticmethod
    def _remove_nones(kwargs):
        return {k: v for k, v in kwargs.items() if v is not None}

    def _get_possible_arguments(self, command=None):
        return super()._get_possible_arguments(command.split()[0])

    def _run(self, cmd, verbosity=True, return_output=False, **kwargs):
        # Constructs the command to run and runs it
        cmd = self._add_arguments(cmd, **kwargs)

        ret = run(cmd, verbose=verbosity, return_output=return_output)
        if isinstance(ret, tuple) and ret[0] != 0:
            _print("WARN: Running command: '%s' failed. Return with output." % cmd)
        elif isinstance(ret, int) and ret != 0:
            _print("WARN: Running command: '%s' failed." % cmd)
        return ret

    def cache_check(self, source_file=None, source_vg=None, source_lv=None, **kwargs):
        cmd = "cache_check "
        if source_file:
            cmd += "%s " % source_file
        if source_vg and source_lv:
            cmd += "%s " % _get_device_path(source_vg, source_lv)
        return self._run(cmd, **self._remove_nones(kwargs))

    def cache_dump(self, source_file=None, source_vg=None, source_lv=None, **kwargs):
        cmd = "cache_dump "
        if source_file:
            cmd += "%s " % source_file
        if source_vg and source_lv:
            cmd += "%s " % _get_device_path(source_vg, source_lv)
        return self._run(cmd, **self._remove_nones(kwargs))

    def cache_metadata_size(self, **kwargs):
        cmd = "cache_metadata_size "
        return self._run(cmd, **self._remove_nones(kwargs))

    def cache_repair(self, **kwargs):
        cmd = "cache_repair "
        return self._run(cmd, **self._remove_nones(kwargs))

    def cache_restore(self, **kwargs):
        cmd = "cache_restore "
        return self._run(cmd, **self._remove_nones(kwargs))

    def cache_writeback(self, **kwargs):
        cmd = "cache_writeback "
        return self._run(cmd, **self._remove_nones(kwargs))

    def thin_check(self, source_file=None, source_vg=None, source_lv=None, **kwargs):
        cmd = "thin_check "
        if source_file:
            cmd += "%s " % source_file
        if source_vg and source_lv:
            cmd += "%s " % _get_device_path(source_vg, source_lv)
        return self._run(cmd, **self._remove_nones(kwargs))

    def thin_delta(self, source_file=None, source_vg=None, source_lv=None, **kwargs):
        cmd = "thin_delta "
        if source_file:
            cmd += "%s " % source_file
        if source_vg and source_lv:
            cmd += "%s " % _get_device_path(source_vg, source_lv)
        return self._run(cmd, **self._remove_nones(kwargs))

    def thin_dump(self, source_file=None, source_vg=None, source_lv=None, **kwargs):
        cmd = "thin_dump "
        if source_file:
            cmd += "%s " % source_file
        if source_vg and source_lv:
            cmd += "%s " % _get_device_path(source_vg, source_lv)
        return self._run(cmd, **self._remove_nones(kwargs))

    def thin_ls(self, source_vg=None, source_lv=None, **kwargs):
        cmd = "thin_ls "
        if source_vg and source_lv:
            cmd += _get_device_path(source_vg, source_lv)
        return self._run(cmd, **self._remove_nones(kwargs))

    def thin_metadata_size(self, **kwargs):
        cmd = "thin_metadata_size "
        return self._run(cmd, **self._remove_nones(kwargs))

    def thin_repair(self, **kwargs):
        cmd = "thin_repair "
        return self._run(cmd, **self._remove_nones(kwargs))

    def thin_restore(self, **kwargs):
        cmd = "thin_restore "
        return self._run(cmd, **self._remove_nones(kwargs))

    def thin_rmap(self, source_file=None, source_vg=None, source_lv=None, **kwargs):
        cmd = "thin_rmap "
        if source_file:
            cmd += "%s " % source_file
        if source_vg and source_lv:
            cmd += "%s " % _get_device_path(source_vg, source_lv)
        return self._run(cmd, **self._remove_nones(kwargs))

    def thin_trim(self, **kwargs):
        cmd = "thin_trim "
        return self._run(cmd, **self._remove_nones(kwargs))



# Copyright (C) 2016 Red Hat, Inc.
# This file is part of libsan.
#
# libsan is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# libsan is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with libsan.  If not, see <http://www.gnu.org/licenses/>.

"""scsi_debug.py: Module to manipulate devices created by scsi_debug module."""

__author__ = "Bruno Goncalves"
__copyright__ = "Copyright (c) 2016 Red Hat, Inc. All rights reserved."

import libsan.host.linux
import libsan.host.mp
import libsan.host.scsi
from libsan import _print
from libsan.host.cmdline import run


def scsi_debug_load_module(options=None):
    module_cmd = "scsi_debug"
    if libsan.host.linux.is_module_loaded(module_cmd):
        _print("WARN: scsi_debug_load_module() - Module is already loaded")
        return True

    if options:
        module_cmd += " %s" % options

    if not libsan.host.linux.load_module(module_cmd):
        _print("FAIL: scsi_debug_load_module() - Could not load %s" % module_cmd)
        return False
    # Wait a bit, for example for multipath to create the device
    libsan.host.linux.sleep(2)
    return True


def scsi_debug_unload_module():
    module_name = "scsi_debug"
    if not libsan.host.linux.is_module_loaded(module_name):
        # Module is not loaded, return success
        return True

    if libsan.host.mp.is_multipathd_running():
        mpaths = libsan.host.mp.mpath_names_of_vendor("Linux")
        for mpath in mpaths:
            libsan.host.mp.remove_mpath(mpath)
            # Wait a bit, for example for multipath to remove the device
            libsan.host.linux.sleep(2)

    if not libsan.host.linux.unload_module(module_name):
        _print("FAIL: scsi_debug_load_module() - Could not unload %s" % module_name)
        return False
    return True


def get_scsi_debug_devices():
    """Return a list of scsi_debug devices"""
    module_name = "scsi_debug"
    vendor = "Linux"
    if not libsan.host.linux.is_module_loaded(module_name):
        return None

    mpaths = libsan.host.mp.mpath_names_of_vendor(vendor)
    if mpaths:
        return mpaths

    scsi_devices = libsan.host.scsi.get_scsi_name_by_vendor(vendor)
    if scsi_devices:
        return scsi_devices

    return None


def scsi_debug_set_param(param, value):
    """Set specific value to scsi debug parameter"""
    if param is None or value is None:
        _print("FAIL: scsi_debug_set_param() - requires param_name and value")
        return False

    if run(f"echo '{value}' > /sys/bus/pseudo/drivers/scsi_debug/{param}", verbose=False) != 0:
        _print(f"FAIL: scsi_debug_set_param() - Could not set {param} with value {value}")
        return False
    return True


def scsi_debug_insert_failure(every_nth, opts):
    """Purpose:
    \tEnable/Disable failure on scsi_debug device
    Parameter:
    \tA dictionary with all option to be set
    \tThe supported paramenters are defined at: http://sg.danny.cz/sg/sdebug26.html
    \topts: 1 - "noisy"
    \t2-"medium error"
    \t4 - ignore "nth"
    \t8 - cause "nth" read or write command to yield a RECOVERED_ERROR
    \t16 -  cause "nth" read or write command to yield a ABORTED_COMMAND
    every_nth: how often the failure is inserted
    Return:
    \tTrue: if success
    \tFalse: if some problem occurred
    """

    if not every_nth:
        every_nth = 0
    if not opts:
        opts = 0

    if not scsi_debug_set_param("every_nth", every_nth):
        _print("FAIL: scsi_debug_insert_failure() - Could not set every_nth with value: %s" % every_nth)
        return False
    if not scsi_debug_set_param("opts", opts):
        _print("FAIL: scsi_debug_insert_failure() - Could not set opts with value: %s" % opts)
        return False
    return True


def self_test():
    """ """
    if not scsi_debug_load_module():
        _print("FAIL: self_test() - Could not load the module")
        return False

    if not get_scsi_debug_devices():
        _print("FAIL: self_test() - Could not find any scsi debug device")
        scsi_debug_unload_module()
        return False

    if not scsi_debug_insert_failure(0, 0):
        _print("FAIL: self_test() - Could not set parameters")
        scsi_debug_unload_module()
        return False

    if not scsi_debug_unload_module():
        _print("FAIL: self_test() - Could not unload the module")
        return False

    return True



#!/usr/bin/python

# Copyright (C) 2016 Red Hat, Inc.
# This file is part of libsan.
#
# libsan is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# libsan is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with libsan.  If not, see <http://www.gnu.org/licenses/>.

from unittest.mock import patch

import pytest
from libsan.host import iscsi

target = "localhost"


def test_install_initiator():
    if not iscsi.install():
        pytest.fail("FAIL: Could not install iSCSI initiator package")
    assert 1


@patch("libsan.host.iscsi.run")
def test_query_discovery(iscsi_run_func):
    discovery_output = """
SENDTARGETS:
DiscoveryAddress: localhost,3260
Target: iqn.2009-10.com.redhat:storage-0
    Portal: [::1]:3260,1
        Iface Name: default
iSNS:
No targets found.
STATIC:
No targets found.
FIRMWARE:
No targets found.
"""
    iscsi_run_func.return_value = (0, discovery_output)

    expected_ret = {
        "SENDTARGETS": {
            "localhost,3260": {
                "disc_addr": "localhost",
                "disc_port": "3260",
                "mode": "sendtargets",
                "targets": {
                    "iqn.2009-10.com.redhat:storage-0": {
                        "portal": {"address": "[::1]", "port": "3260"},
                        "iface": ["default"],
                    }
                },
            }
        },
        "iSNS": {},
        "STATIC": {},
        "FIRMWARE": {},
    }
    if iscsi.query_discovery() != expected_ret:
        pytest.fail("FAIL: Could not  query discovery iSCSI target")
    assert 1


@patch("libsan.host.iscsi.run")
def test_discovery(iscsi_run_func):
    discovery_output = "[::1]:3260,1 iqn.2009-10.com.redhat:storage-0"
    iscsi_run_func.return_value = (0, discovery_output)

    if not iscsi.discovery_st(target):
        pytest.fail("FAIL: Could not discovery iSCSI target")
    assert 1


@patch("libsan.host.linux.service_restart")
def test_set_iscsid_parameter(service_restart_func):
    service_restart_func.return_value = True
    if not iscsi.set_iscsid_parameter({"node.session.cmds_max": "4096", "node.session.queue_depth": "128"}):
        pytest.fail("FAIL: Unable to set iscsid parameter")
    assert 1
